# -*- coding: utf-8 -*-
"""
Created by gregory on 18.03.16

Copyright 2016 Alpes Lasers SA, Neuchatel, Switzerland
"""
import csv
import json
import hashlib
from os import path
import logging

from .vcm_calibration import (validate_calibration,
                                                       angle_for_wavelength as _angle_for_wavelength,
                                                       wavelength_for_angle as _wavelength_for_angle)
from .vcm_calibration import (validate_calibration,
                                                       angle_for_wavelength_old as _angle_for_wavelength_old,
                                                       wavelength_for_angle_old as _wavelength_for_angle_old)
from .file_backed_dict import FileBackedDict

logger = logging.getLogger(__name__)
__author__ = 'gregory'
__copyright__ = "Copyright 2016, Alpes Lasers SA"

# Configuration keys.
DEBUG = "DEBUG"
SERVER_PORT = "SERVER_PORT"
CERTIFICATE_PATH = "CERTIFICATE_PATH"
PRIVATE_KEY_PATH = "PRIVATE_KEY_PATH"
CONFIGURATION_FILE_PATH = "CONFIGURATION_FILE_PATH"
USERS_FILE_PATH = "USERS_FILE_PATH"
CALIBRATION_FILE_PATH = "CALIBRATION_FILE_PATH"
ADMINS_FILE_PATH = "ADMINS_FILE_PATH"
PULSER_PORT = "PULSER_PORT"
STAGE_MODEL = "STAGE_MODEL"
STAGE_SERIAL_NUMBER = "STAGE_SERIAL_NUMBER"
STAGE_PORT = "STAGE_PORT"
STAGE_PROP_FACTOR = "STAGE_PROP_FACTOR"
STAGE_CONTROL_FREQUENCY = "STAGE_CONTROL_FREQUENCY"
STAGE_EXCITATION_FREQUENCY = "STAGE_EXCITATION_FREQUENCY"
STAGE_ANGLE = "STAGE_ANGLE"
STAGE_ANGLE_Y = "STAGE_ANGLE_Y"
TC_SERIAL_PORT = "TC_SERIAL_PORT"
TC_BAUDRATE = "TC_BAUDRATE"
TC_MODEL = "TC_MODEL"
PULSER_GENERATION = "PULSER_GENERATION"
PULSER_MINIMUM_VOLTAGE = "PULSER_MINIMUM_VOLTAGE"
PULSER_MAXIMUM_VOLTAGE = "PULSER_MAXIMUM_VOLTAGE"
PULSER_MAXIMUM_CURRENT = "PULSER_MAXIMUM_CURRENT"
PULSER_MINIMUM_PULSE_PERIOD = "PULSER_MINIMUM_PULSE_PERIOD"
PULSER_MAXIMUM_PULSE_PERIOD = "PULSER_MAXIMUM_PULSE_PERIOD"
PULSER_MINIMUM_PULSE_WIDTH = "PULSER_MINIMUM_PULSE_WIDTH"
PULSER_MAXIMUM_PULSE_WIDTH = "PULSER_MAXIMUM_PULSE_WIDTH"
MINIMUM_DUTY_CYCLE = "MINIMUM_DUTY_CYCLE"
MAXIMUM_DUTY_CYCLE = "MAXIMUM_DUTY_CYCLE"
MAXIMUM_MIRROR_AMPLITUDE= "MAXIMUM_MIRROR_AMPLITUDE"
BYPASS_TC_CHECK = "BYPASS_TC_CHECK"
MINIMUM_STAGE_VELOCITY = "MINIMUM_STAGE_VELOCITY"
MAXIMUM_STAGE_VELOCITY = "MAXIMUM_STAGE_VELOCITY"
ANGLE_UNITS = "ANGLE_UNITS"
WAVELENGTH_UNITS = "WAVELENGTH_UNITS"

TEC_TEMPERATURE = 'temperature'
TEC_TEMPERATURE_CHANGE_RATE = 'temperature_change_rate'
TEC_INVERSE_POLARITY = 'inverse_polarity'
TEC_TEMPERATURE_TOLERANCE = 'temperature_tolerance'
TEC_MIN_TIME_STABLE = 'min_time_stable'
TEC_CONTROL_P = 'control_p'
TEC_CONTROL_I = 'control_i'
TEC_CONTROL_D = 'control_d'
TEC_MAX_VOLTAGE = 'max_voltage'
TEC_MAX_CURRENT='max_current'
TEC_NTC_RESISTANCE_0 = 'resistance_0'
TEC_NTC_RESISTANCE_25 = 'resistance_25'
TEC_NTC_RESISTANCE_60 = 'resistance_60'

PELTIER_MAX_CURRENT = 'max_current'
PELTIER_DELTA_T = 'delta_t'


configuration_dictionary = {
  "ADMINS_FILE_PATH": "admins",
  "ANGLE_UNITS": "deg",
  "BYPASS_TC_CHECK": False,
  "CALIBRATION_FILE_PATH": "/home/pi/ec_config/calibration1",
  "CERTIFICATE_PATH": "",
  "CONFIGURATION_FILE_PATH": "/home/pi/ec_config/config1.json",
  "DEBUG": False,
  "MAXIMUM_DUTY_CYCLE": 0.05,
  "MAXIMUM_MIRROR_AMPLITUDE": 25,
  "MAXIMUM_STAGE_VELOCITY": 180,
  "MINIMUM_DUTY_CYCLE": 0,
  "MINIMUM_STAGE_VELOCITY": 0.05,
  "PRIVATE_KEY_PATH": "",
  "PULSER_GENERATION": 5,
  "PULSER_MAXIMUM_CURRENT": 3.0,
  "PULSER_MAXIMUM_PULSE_PERIOD": 1310720,
  "PULSER_MAXIMUM_PULSE_WIDTH": 1310720,
  "PULSER_MAXIMUM_VOLTAGE": 0.6,
  "PULSER_MINIMUM_PULSE_PERIOD": 500,
  "PULSER_MINIMUM_PULSE_WIDTH": 200,
  "PULSER_MINIMUM_VOLTAGE": 0,
  "PULSER_PORT": "/dev/tty.ec_pulser1",
  "SERVER_PORT": 5000,
  "STAGE_ANGLE": -0.00011753971488422282,
  "STAGE_ANGLE_Y": 0,
  "STAGE_CONTROL_FREQUENCY": 1000,
  "STAGE_EXCITATION_FREQUENCY": 171500,
  "STAGE_MODEL": "MR-E",
  "STAGE_PORT": "/dev/tty.ec_vcm1",
  "STAGE_PROP_FACTOR": 60,
  "STAGE_SERIAL_NUMBER": "3",
  "STYLE": {'background': '8cff96',
            'header': 'test'},
  "TC_BAUDRATE": 38400,
  "TC_MODEL": "MEERS",
  "TC_SERIAL_PORT": "/dev/tty.ec_tc_meer1",
  "USERS_FILE_PATH": "users",
  "WAVELENGTH_UNITS": "nm"
}



def _get_subconfig_filepath(file_key):
    dirpath, _ = path.split(configuration_dictionary.get(CONFIGURATION_FILE_PATH, ''))
    filepath = path.join(dirpath, configuration_dictionary.get(file_key))
    return filepath


admins = FileBackedDict(_get_subconfig_filepath(ADMINS_FILE_PATH))
users = FileBackedDict(_get_subconfig_filepath(USERS_FILE_PATH))
_calibration = None
config_backup_operations = 100
config_write_operations = config_backup_operations


def load_configuration(filepath):
    global admins
    global users
    global _calibration

    try:
        with open(filepath) as fp:
            new_config = json.load(fp)
    except Exception:
        with open('.' + filepath + '.bk') as fp:
            new_config = json.load(fp)

    configuration_dictionary.update(new_config, **{CONFIGURATION_FILE_PATH: filepath})
    admins.load(_get_subconfig_filepath(ADMINS_FILE_PATH), clear=True)
    users.load(_get_subconfig_filepath(USERS_FILE_PATH), clear=True)
    _calibration = None
    get_calibration()


def update_configuration(config_dict=(), **kwargs):
    global admins
    global users
    global _calibration
    global config_write_operations
    new = False
    for k in config_dict:
        if config_dict[k] != configuration_dictionary.get(k):
            new = True
            break
    if not new:
        for k in kwargs:
            if kwargs[k] != configuration_dictionary.get(k):
                new = True
                break
    configuration_dictionary.update(config_dict, **kwargs)
    if ADMINS_FILE_PATH in config_dict or ADMINS_FILE_PATH in kwargs:
        admins.load(_get_subconfig_filepath(ADMINS_FILE_PATH), clear=True)
    if USERS_FILE_PATH in config_dict or USERS_FILE_PATH in kwargs:
        users.load(_get_subconfig_filepath(USERS_FILE_PATH), clear=True)
    if CALIBRATION_FILE_PATH in config_dict or CALIBRATION_FILE_PATH in kwargs:
        _calibration = None
        get_calibration()

    # Save if there is an available CONFIGURATION_FILE_PATH.
    filepath = configuration_dictionary.get(CONFIGURATION_FILE_PATH)
    if filepath and new:
        with open(filepath, 'w') as fp:
            json.dump(configuration_dictionary, fp, indent=2, sort_keys=True)
        config_write_operations += 1
        if config_write_operations >= config_backup_operations:
            config_write_operations = 0
            with open(filepath + '.bk', 'w') as fp:
                json.dump(configuration_dictionary, fp, indent=2, sort_keys=True)


def hash_pw(username, password):
    return hashlib.md5((password + str(len(username)) + username + str(len(password)) + '#*salty').encode()).hexdigest()


def get_calibration():
    global _calibration
    if _calibration:
        return _calibration
    try:
        with open(_get_subconfig_filepath(CALIBRATION_FILE_PATH)) as fp:
            _calibration = json.load(fp)
    except IOError:
        return []
    return _calibration


def set_calibration(calibration):
    global _calibration
    validate_calibration(calibration)
    _calibration = calibration
    filepath = _get_subconfig_filepath(CALIBRATION_FILE_PATH)
    if filepath:
        with open(filepath, 'w') as fp:
            json.dump(calibration, fp, indent=2)


def get_angle_for_wavelength(wavelength):
    """Raises a ValueError if the wavelength is out of bounds."""
    return _angle_for_wavelength(_calibration or [], wavelength)


def get_wavelength_for_angle(angle, out_of_range_exc=True):
    """Raises a ValueError if the wavelength is out of bounds."""
    try:
        return _wavelength_for_angle(_calibration or [], angle)
    except ValueError:
        if not out_of_range_exc:
            return None
        raise


def get_angle_for_wavelength_old(wavelength):
    """Raises a ValueError if the wavelength is out of bounds."""
    return _angle_for_wavelength_old(_calibration or [], wavelength)


def get_wavelength_for_angle_old(angle, out_of_range_exc=True):
    """Raises a ValueError if the wavelength is out of bounds."""
    try:
        return _wavelength_for_angle_old(_calibration or [], angle)
    except ValueError:
        if not out_of_range_exc:
            return None
        raise


if __name__ == '__main__':
    calibration = [(320, 1000), (10, 1050)]
    set_calibration(calibration)
    print(get_angle_for_wavelength(1039))
    print(get_wavelength_for_angle(321))



# -*- coding: utf-8 -*-
"""
Created by chiesa

Copyright Alpes Lasers SA, Switzerland
"""
__author__ = 'chiesa'
__copyright__ = "Copyright Alpes Lasers SA"

from argparse import ArgumentParser
from time import sleep

from phootonics_controller.base_controllers.stage_vcm import VCMirrorController

if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument('-p', type=str, required=True)
    parser.add_argument('-a', type=float, required=True)
    args = parser.parse_args()
    vcmc = VCMirrorController(args.mirror)
    vcmc.start()
    while True:
        if vcmc.status == "READY":
            break
        sleep(0.2)
    vcmc.move_to_angle(args.a)
    print(vcmc.get_angle())
    vcmc.stop()




# -*- coding: utf-8 -*-
"""
Created by chiesa

Copyright Alpes Lasers SA, Switzerland
"""
__author__ = 'chiesa'
__copyright__ = "Copyright Alpes Lasers SA"

import itertools
from argparse import ArgumentParser
import logging
from time import sleep

import requests

import pandas

from phootonics_controller.base_controllers.config import WL_INCREMENT

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

PHOOTONICS_URL = 'http://0.0.0.0:5000'


def scan():
    parser = ArgumentParser()
    parser.add_argument('-x', nargs='+', type=float)
    parser.add_argument('-y', nargs='+', type=float)
    parser.add_argument('-ampl_x', type=float, default=5)
    parser.add_argument('-ampl_y', type=float, default=5)
    parser.add_argument('-wl_step', type=float, required=False, default=WL_INCREMENT)
    parser.add_argument('-scan', type=str, choices=['default', 'fast'], required=True)
    parser.add_argument('-url', type=str, required=False, default=PHOOTONICS_URL)
    args=parser.parse_args()


    rsp = requests.get(args.url + '/info')
    rsp.raise_for_status()
    info = rsp.json()
    if info['scan_running']:
        parser.error('Scan command already running.')


    if args.scan == 'default':
        x_list = args.x
        y_list = args.y

        positions = list(itertools.product(x_list, y_list))

        rsp = requests.post(args.url + '/scan/start/default',
                            json={'positions': positions, 'wl_step': args.wl_step})
        if not rsp.status_code == 200:
            print(rsp.json())
            rsp.raise_for_status()
    elif args.scan == 'fast':
        rsp = requests.post(args.url + '/scan/start/fast',
                            json={'wl_step': args.wl_step,
                                  'amplitude_x': args.ampl_x,
                                  'amplitude_y': args.ampl_y})
        if not rsp.status_code == 200:
            print(rsp.json())
            rsp.raise_for_status()

    while True:
        sleep(1)
        rsp = requests.get(args.url + '/info')
        rsp.raise_for_status()
        info = rsp.json()
        print('Completed: {:.0%}'.format(info['scan']['runtime']['completed']))
        if not info['scan_running']:
            break

    rsp = requests.get(args.url + '/info')
    rsp.raise_for_status()
    info = rsp.json()

    df = pandas.DataFrame.from_records(info['scan']['results'],
                                       columns=['active_cavity',
                                                'detector_voltage',
                                                'wavelength',
                                                'wavelength_set_point',
                                                'x',
                                                'x_set_point',
                                                'y',
                                                'y_set_point'])
    df.to_csv('outputfile.csv')

if __name__ == '__main__':
    scan()








from serial.tools import list_ports

from .SPIDevice import SPI
import sys, os
import time

class MR1030:
    _valid_hwids = ['0483:A31E']
    _port = ''
    _endian = '>'
    _type_int = 'I'
    _type_flt = 'f'
    
    sysclk = 18000000
    clkdiv = 16
    def __init__(self, bus, device, freq0, amp0, freq1, amp1, iters):
        _int = self._endian + self._type_int
        _flt = self._endian + self._type_flt
        
        self.spi = SPI(bus=bus, device=device)
        self.spi._spi_comm.max_speed_hz = self.sysclk//self.clkdiv
        self.filename = 'point_cloud.txt'

        #### Set current example
        #ans = self.spi.set_values(0x50,0x00, 0x51, 0x00,0.05 , -0.08 , _flt) 
        #print(ans)

        #### Set analog mode example
        #ans = self.spi.set_values(0x40,0x00, 0x40, 0x05, 0x58, 0x59, _int) 
        #print(ans)

        ### Resonant mirror example
        ans = self.spi.set_values(0x40, 0x00, 0x40, 0x05, 0x60, 0x61, _int)   # Signal-Gen set as input
        print(ans)        
        
        ans = self.spi.set_values(0x40,0x02, 0x40, 0x07, 0xC0, 0xC1, _int)   # both axis closed loop
        #ans = self.spi.set_values(0x40,0x02, 0x40, 0x07, 0xC0, 0xB1, _int)   # X axis closed loop, Y axis open loop 
        print(ans)
        
        ans = self.spi.set_values(0x60, 0x00, 0x61, 0x00, 2, 2, _int)         # Signal-Gen Unit
        print(ans)
        
        ans = self.spi.set_values(0x60, 0x02, 0x61, 0x02, 1, 0, _int)         # Signal-Gen Shape
        print(ans)        
        
        ans = self.spi.set_values(0x60, 0x03, 0x61, 0x03, freq0, freq1, _flt) # Signal-Gen Frequency
        print(ans)
        
        ans = self.spi.set_values(0x60, 0x04, 0x61, 0x04, amp0, amp1, _flt)   # Signal-Gen Amplitude
        print(ans)
        
        ans = self.spi.set_values(0x60, 0x01, 0x61, 0x01, 1, 1, _int)         # Signal-Gen Run
        print(ans)
        
        ans = self.spi._spi_comm.xfer([0, 0,  0x30, 0x02, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        print([hex(i) for i in ans])

        time.sleep(2)
        #self.listen_raw(iters)  # only return raw values, not struct.unpacked
        self.listen(iters)  # unpack values
              
        ans = self.spi.set_values(0x60, 0x01, 0x61, 0x01, 0, 0, _int)  # Signal-Gen Stop
        ans = self.spi._spi_comm.xfer([0, 0, 0x22, 0x00, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        print([hex(i) for i in ans])

        ans = self.spi._spi_comm.xfer([0, 0, 0x22, 0x00, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        print([hex(i) for i in ans])
        
    def listen_raw(self, iterations):
        start = time.time()
        with open(self.filename, 'w') as file:
            self.spi._spi_comm.xfer([0, 0, 0x30, 0x02, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
            for _ in range (iterations):
                res = self.spi._spi_comm.xfer([0, 0, 0x30, 0x02, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
                file.write(str([hex(i) for i in res])+'\n')
            print("Time Taken: {}s".format(time.time()-start))
            file.close()
            
    def listen(self, iterations):
        start = time.time()
        with open(self.filename, 'w') as file:
            self.spi._spi_comm.xfer([0, 0, 0x30, 0x02, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
            for _ in range (iterations):
                reg, x, y = self.spi.get_values(0x30, 0x02, '>I')
                file.write(str(reg)+','+str(x)+','+str(y)+'\n')
            print("Time Taken: {}s".format(time.time()-start))
            file.close()
        
if __name__ == '__main__':
    mr1030 = MR1030(bus=0, device=0, freq0=5.0, amp0=0.4, freq1=1.0, amp1=0.05, iters=10000)



# -*- coding: utf-8 -*-
from external_cavity.controllers.driver_utils.MR_E_2_Python_SDK.optoMDC.mre2 import MRE2Board
mre2 = MRE2Board(port='/dev/ttyACM0', verbose=True)  # this will connect and synchronize all channel with the firmware
mirror = mre2.Mirror
from time import sleep
from external_cavity.controllers.driver_utils.MR_E_2_Python_SDK.optoKummenberg.tools.systems_registers_tools import get_registers
import math

# TODO: write initialisation function trying to reconnect and reinitializing, busy while performing commands

def set_xy_control_mode_both_axes():
    current_control_mode_x = mirror.Channel_0.GetControlMode()
    if current_control_mode_x != 2:
        mirror.Channel_0.SetControlMode(2)
    current_control_mode_y = mirror.Channel_1.GetControlMode()
    if current_control_mode_y != 2:
        mirror.Channel_1.SetControlMode(2)


def get_current_xy_position():
    current_position_x = mirror.Channel_0.StaticInput.GetXY()
    current_position_y = mirror.Channel_1.StaticInput.GetXY()
    return current_position_x[0], current_position_y[0]


def move_to_xy_position(posx, posy):
        if posx**2+posy**2 <= 1.0:
            mirror.Channel_0.StaticInput.SetXY(posx)
            mirror.Channel_1.StaticInput.SetXY(posy)
        else:
            Exception ('XY position out of range, x² + y² should be <= 1.0')


def move_along_x_stepwise_to_position(target_x, step, delay):
        x, y = get_current_xy_position()
        if target_x < x:
            step_x = -step
        else:
            step_x = step
        distance_x = abs(target_x - x)
        x_0 = 0
        while x_0 + abs(step_x) <= distance_x:
            x, y = get_current_xy_position()
            move_to_xy_position(x + step_x, y)
            sleep(delay)
            xcurr, ycurr = get_current_xy_position()
            x_0 = x_0 + abs(step_x)
            print(xcurr, ycurr)


def move_along_x_stepwise_from_a_to_b(posx_a, posx_b, step, delay):
    mirror.Channel_0.StaticInput.SetAsInput()
    x, y = get_current_xy_position()
    move_to_xy_position(posx_a,y)
    if posx_b < posx_a:
        step_x = -step
    else:
        step_x = step
    distance_x = abs(posx_b-posx_a)
    x_0 = 0
    while x_0 + abs(step_x) <= distance_x:
        x, y = get_current_xy_position()
        move_to_xy_position(x + step_x, y)
        sleep(delay)
        xcurr, ycurr = get_current_xy_position()
        x_0 = x_0 + abs(step_x)
        print(xcurr, ycurr)


def move_along_y_stepwise_to_position(target_y, step, delay):
        x, y = get_current_xy_position()
        if target_y < y:
            step_y = -step
        else:
            step_y = step
        distance_y = abs(target_y - y)
        y_0 = 0
        while y_0 + abs(step_y) <= distance_y:
            x, y = get_current_xy_position()
            move_to_xy_position(x, y + step_y)
            sleep(delay)
            xcurr, ycurr = get_current_xy_position()
            y_0 = y_0 + abs(step_y)
            print(xcurr, ycurr)


def sweep_along_x_from_a_to_b(start_angle, stop_angle, speed, n_repetitions):
    start_position = convert_angles_deg_to_position_x(start_angle)
    print(start_position)
    stop_position= convert_angles_deg_to_position_x(stop_angle)
    print(stop_position)
    speed_conv=convert_angles_deg_to_position_x(speed)
    mirror.Channel_0.StaticInput.SetXY(start_position)
    sleep(5)
    sig_gen = mre2.Mirror.Channel_0.SignalGenerator
    sig_gen.SetAsInput()
    frequency =float(1/(abs(stop_position-start_position)/speed_conv))
    sig_gen.SetUnit(2)
    sig_gen.SetShape(0)
    sig_gen.SetFrequency(frequency)
    amplitude = float(abs(stop_position-start_position))
    sig_gen.SetAmplitude(amplitude)
    sig_gen.SetCycles(n_repetitions)
    sig_gen.Run()


def convert_position_to_angles_xy_deg(posx, posy):
    tan_50=math.tan(math.radians(50))
    angle_x= math.degrees(math.atan(posx * tan_50))
    angle_y = math.degrees(math.atan(posy * tan_50))
    return angle_x, angle_y


def convert_angles_deg_to_position_xy(degx, degy):
    tan_50=math.tan(math.radians(50))
    posx = math.tan(math.radians(degx))/tan_50
    posy = math.tan(math.radians(degy))/tan_50
    return posx, posy

def convert_x_position_to_angles_deg(pos):
    tan_50 = math.tan(math.radians(50))
    angle = math.degrees(math.atan(pos * tan_50))
    return angle

def convert_angles_deg_to_position_x(angle):
    tan_50 = math.tan(math.radians(50))
    pos = math.tan(math.radians(angle)) / tan_50
    return pos


def is_not_out_of_bounds(position_x):
    y= mre2.Mirror.OpticalFeedback.get_register('y')[0]
    if (position_x**2 + y**2)<= 1.0:
        return True
    else:
        return False


def move_rel(increment):
    current_position_x = float(mre2.Mirror.OpticalFeedback.get_register('x')[0])
    current_angle = convert_x_position_to_angles_deg(current_position_x)
    target_angle = current_angle + increment
    target_position= convert_angles_deg_to_position_x(target_angle)
    if is_not_out_of_bounds(target_position):
        mre2.Mirror.Channel_0.StaticInput.SetXY(target_position)
    else:
        print('out of bounds')


if __name__ == '__main__':
    start_position = convert_angles_deg_to_position_x(0)
    print(start_position)
    end_position = convert_angles_deg_to_position_x(6)
    print(end_position)
    conv_velocity = convert_angles_deg_to_position_x(0.5)
    frequency = float(1 / ((abs(end_position - start_position) / conv_velocity)))
    print(frequency)
    amplitude = float(abs(end_position - start_position)) /2
    print(amplitude)
    offset = (1/frequency)/4
    print(offset)







import logging

from phootonics_controller.utils.ADC_converter import ADS1263

ADC = ADS1263.ADS1263()
ADC_CHANNEL = 0
REF = 5.08


def initialise_adc():
    if ADC.ADS1263_init() == -1:
        raise Exception('ADC not ininitalized')


def read_adc_detector():
    try:
        adc_value = ADC.ADS1263_GetAll()
        print(adc_value)
        if adc_value[ADC_CHANNEL] >> 31 == 1:
            value = REF * 2 - adc_value[ADC_CHANNEL] * REF / 0x80000000
        else:
            value = adc_value[ADC_CHANNEL] * REF / 0x7fffffff  # 32bit
        return value
    except Exception as e:
        logging.error('could not read ADC value {}'.format(e))
        ADC.ADS1263_Exit()



if __name__ == '__main__':
    initialise_adc()
    value = read_adc_detector()
    print(value)


# -*- coding: utf-8 -*-
"""
Created by olgare on 10.09.2021

Copyright 2021 Alpes Lasers SA, Neuchatel, Switzerland
"""
__author__ = 'olgare'
__copyright__ = "Copyright 2021, Alpes Lasers SA"

from time import sleep

import requests
url = 'http://localhost:5555'
if __name__ == '__main__':
    rsp = requests.get(url + '/info')

    rsp.raise_for_status()
    print(rsp.json())
    rsp = requests.post(url + '/scan/start',
                        json={'positions': [(1.0, 1.0), (3.0, 3.0)]})
    if not rsp.status_code == 200:
        print(rsp.json())
        rsp.raise_for_status()
    rsp = requests.get(url + '/info')

    rsp.raise_for_status()
    #rsp = requests.post(url + '/scan/stop')
    rsp.raise_for_status()
    rsp = requests.get(url + '/info')
    print(rsp.json())
    sleep(3.0)
    rsp = requests.get(url + '/info')

    rsp.raise_for_status()
    print(rsp.json())





# -*- coding: utf-8 -*-
"""
Created by gregory on 18.03.16

Copyright 2016 Alpes Lasers SA, Neuchatel, Switzerland
"""
import json
import logging

logger = logging.getLogger(__name__)

__author__ = 'gregory'
__copyright__ = "Copyright 2016, Alpes Lasers SA"


class FileBackedDict(dict):
    """Dictionary that automatically saves the changes to a file"""

    def __init__(self, filepath):
        super(FileBackedDict, self).__init__()
        self.filepath = filepath
        if self.filepath is not None:
            try:
                self.load()
            except IOError:
                pass  # not an error if the file does not exist at creation time.

    def load(self, filepath=None, clear=False):
        if filepath is not None:
            self.filepath = filepath
        try:
            with open(self.filepath) as fp:
                dct = json.load(fp)
        except IOError:
            dct = {}
        if clear:
            self.clear()
        self.update(dct)

    def save(self, filepath=None):
        if filepath is not None:
            self.filepath = filepath
        if self.filepath is not None:
            with open(self.filepath, 'w') as fp:
                json.dump(self, fp)

    def update(self, E=None, **F):
        super(FileBackedDict, self).update(E, **F)
        self.save()

    def setdefault(self, k, d=None):
        super(FileBackedDict, self).setdefault(k, d)
        self.save()

    def __setitem__(self, key, value):
        super(FileBackedDict, self).__setitem__(key, value)
        self.save()

    def __delitem__(self, key):
        super(FileBackedDict, self).__delitem__(key)
        self.save()

    def clear(self):
        super(FileBackedDict, self).clear()
        self.save()



# /*****************************************************************************
# * | File        :	  ADS1263.py
# * | Author      :   Waveshare team
# * | Function    :   ADS1263 driver
# * | Info        :
# *----------------
# * | This version:   V1.0
# * | Date        :   2020-12-15
# * | Info        :
# ******************************************************************************/
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documnetation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to  whom the Software is
# furished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS OR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import phootonics_controller.utils.ADC_converter.ads_config as config
import RPi.GPIO as GPIO

ScanMode = 0

# gain
ADS1263_GAIN = {
    'ADS1263_GAIN_1': 0,  # GAIN   1
    'ADS1263_GAIN_2': 1,  # GAIN   2
    'ADS1263_GAIN_4': 2,  # GAIN   4
    'ADS1263_GAIN_8': 3,  # GAIN   8
    'ADS1263_GAIN_16': 4,  # GAIN  16
    'ADS1263_GAIN_32': 5,  # GAIN  32
    'ADS1263_GAIN_64': 6,  # GAIN  64
}
# ADC2 gain
ADS1263_ADC2_GAIN = {
    'ADS1263_ADC2_GAIN_1': 0,  # GAIN  1
    'ADS1263_ADC2_GAIN_2': 1,  # GAIN  2
    'ADS1263_ADC2_GAIN_4': 2,  # GAIN  4
    'ADS1263_ADC2_GAIN_8': 3,  # GAIN  8
    'ADS1263_ADC2_GAIN_16': 4,  # GAIN  16
    'ADS1263_ADC2_GAIN_32': 5,  # GAIN  32
    'ADS1263_ADC2_GAIN_64': 6,  # GAIN  64
    'ADS1263_ADC2_GAIN_128': 7,  # GAIN  128
}
# data rate
ADS1263_DRATE = {
    'ADS1263_38400SPS': 0xF,
    'ADS1263_19200SPS': 0xE,
    'ADS1263_14400SPS': 0xD,
    'ADS1263_7200SPS': 0xC,
    'ADS1263_4800SPS': 0xB,
    'ADS1263_2400SPS': 0xA,
    'ADS1263_1200SPS': 0x9,
    'ADS1263_400SPS': 0x8,
    'ADS1263_100SPS': 0x7,
    'ADS1263_60SPS': 0x6,
    'ADS1263_50SPS': 0x5,
    'ADS1263_20SPS': 0x4,
    'ADS1263_16d6SPS': 0x3,
    'ADS1263_10SPS': 0x2,
    'ADS1263_5SPS': 0x1,
    'ADS1263_2d5SPS': 0x0,
}
# ADC2 data rate
ADS1263_ADC2_DRATE = {
    'ADS1263_ADC2_10SPS': 0,
    'ADS1263_ADC2_100SPS': 1,
    'ADS1263_ADC2_400SPS': 2,
    'ADS1263_ADC2_800SPS': 3,
}
# Delay time
ADS1263_DELAY = {
    'ADS1263_DELAY_0s': 0,
    'ADS1263_DELAY_8d7us': 1,
    'ADS1263_DELAY_17us': 2,
    'ADS1263_DELAY_35us': 3,
    'ADS1263_DELAY_169us': 4,
    'ADS1263_DELAY_139us': 5,
    'ADS1263_DELAY_278us': 6,
    'ADS1263_DELAY_555us': 7,
    'ADS1263_DELAY_1d1ms': 8,
    'ADS1263_DELAY_2d2ms': 9,
    'ADS1263_DELAY_4d4ms': 10,
    'ADS1263_DELAY_8d8ms': 11,
}
# DAC out volt
ADS1263_DAC_VOLT = {
    'ADS1263_DAC_VLOT_4_5': 0b01001,  # 4.5V
    'ADS1263_DAC_VLOT_3_5': 0b01000,
    'ADS1263_DAC_VLOT_3': 0b00111,
    'ADS1263_DAC_VLOT_2_75': 0b00110,
    'ADS1263_DAC_VLOT_2_625': 0b00101,
    'ADS1263_DAC_VLOT_2_5625': 0b00100,
    'ADS1263_DAC_VLOT_2_53125': 0b00011,
    'ADS1263_DAC_VLOT_2_515625': 0b00010,
    'ADS1263_DAC_VLOT_2_5078125': 0b00001,
    'ADS1263_DAC_VLOT_2_5': 0b00000,
    'ADS1263_DAC_VLOT_2_4921875': 0b10001,
    'ADS1263_DAC_VLOT_2_484375': 0b10010,
    'ADS1263_DAC_VLOT_2_46875': 0b10011,
    'ADS1263_DAC_VLOT_2_4375': 0b10100,
    'ADS1263_DAC_VLOT_2_375': 0b10101,
    'ADS1263_DAC_VLOT_2_25': 0b10110,
    'ADS1263_DAC_VLOT_2': 0b10111,
    'ADS1263_DAC_VLOT_1_5': 0b11000,
    'ADS1263_DAC_VLOT_0_5': 0b11001,
}
# registration definition
ADS1263_REG = {
    # Register address, followed by reset the default values
    'REG_ID': 0,  # xxh
    'REG_POWER': 1,  # 11h
    'REG_INTERFACE': 2,  # 05h
    'REG_MODE0': 3,  # 00h
    'REG_MODE1': 4,  # 80h
    'REG_MODE2': 5,  # 04h
    'REG_INPMUX': 6,  # 01h
    'REG_OFCAL0': 7,  # 00h
    'REG_OFCAL1': 8,  # 00h
    'REG_OFCAL2': 9,  # 00h
    'REG_FSCAL0': 10,  # 00h
    'REG_FSCAL1': 11,  # 00h
    'REG_FSCAL2': 12,  # 40h
    'REG_IDACMUX': 13,  # BBh
    'REG_IDACMAG': 14,  # 00h
    'REG_REFMUX': 15,  # 00h
    'REG_TDACP': 16,  # 00h
    'REG_TDACN': 17,  # 00h
    'REG_GPIOCON': 18,  # 00h
    'REG_GPIODIR': 19,  # 00h
    'REG_GPIODAT': 20,  # 00h
    'REG_ADC2CFG': 21,  # 00h
    'REG_ADC2MUX': 22,  # 01h
    'REG_ADC2OFC0': 23,  # 00h
    'REG_ADC2OFC1': 24,  # 00h
    'REG_ADC2FSC0': 25,  # 00h
    'REG_ADC2FSC1': 26,  # 40h
}
# comand
ADS1263_CMD = {
    'CMD_RESET': 0x06,  # Reset the ADC, 0000 011x (06h or 07h)
    'CMD_START1': 0x08,  # Start ADC1 conversions, 0000 100x (08h or 09h)
    'CMD_STOP1': 0x0A,  # Stop ADC1 conversions, 0000 101x (0Ah or 0Bh)
    'CMD_START2': 0x0C,  # Start ADC2 conversions, 0000 110x (0Ch or 0Dh)
    'CMD_STOP2': 0x0E,  # Stop ADC2 conversions, 0000 111x (0Eh or 0Fh)
    'CMD_RDATA1': 0x12,  # Read ADC1 data, 0001 001x (12h or 13h)
    'CMD_RDATA2': 0x14,  # Read ADC2 data, 0001 010x (14h or 15h)
    'CMD_SYOCAL1': 0x16,  # ADC1 system offset calibration, 0001 0110 (16h)
    'CMD_SYGCAL1': 0x17,  # ADC1 system gain calibration, 0001 0111 (17h)
    'CMD_SFOCAL1': 0x19,  # ADC1 self offset calibration, 0001 1001 (19h)
    'CMD_SYOCAL2': 0x1B,  # ADC2 system offset calibration, 0001 1011 (1Bh)
    'CMD_SYGCAL2': 0x1C,  # ADC2 system gain calibration, 0001 1100 (1Ch)
    'CMD_SFOCAL2': 0x1E,  # ADC2 self offset calibration, 0001 1110 (1Eh)
    'CMD_RREG': 0x20,  # Read registers 001r rrrr (20h+000r rrrr)
    'CMD_RREG2': 0x00,  # number of registers to read minus 1, 000n nnnn
    'CMD_WREG': 0x40,  # Write registers 010r rrrr (40h+000r rrrr)
    'CMD_WREG2': 0x00,  # number of registers to write minus 1, 000n nnnn
}


class ADS1263:
    def __init__(self):
        self.rst_pin = config.RST_PIN
        self.cs_pin = config.CS_PIN
        self.drdy_pin = config.DRDY_PIN

    # Hardware reset
    def ADS1263_reset(self):
        config.digital_write(self.rst_pin, GPIO.HIGH)
        config.delay_ms(200)
        config.digital_write(self.rst_pin, GPIO.LOW)
        config.delay_ms(200)
        config.digital_write(self.rst_pin, GPIO.HIGH)
        config.delay_ms(200)

    def ADS1263_WriteCmd(self, reg):
        config.digital_write(self.cs_pin, GPIO.LOW)  # cs  0
        config.spi_writebyte([reg])
        config.digital_write(self.cs_pin, GPIO.HIGH)  # cs 1

    def ADS1263_WriteReg(self, reg, data):
        config.digital_write(self.cs_pin, GPIO.LOW)  # cs  0
        config.spi_writebyte([ADS1263_CMD['CMD_WREG'] | reg, 0x00, data])
        config.digital_write(self.cs_pin, GPIO.HIGH)  # cs 1

    def ADS1263_ReadData(self, reg):
        config.digital_write(self.cs_pin, GPIO.LOW)  # cs  0
        config.spi_writebyte([ADS1263_CMD['CMD_RREG'] | reg, 0x00])
        data = config.spi_readbytes(1)
        config.digital_write(self.cs_pin, GPIO.HIGH)  # cs 1
        return data

    # Check Data
    def ADS1263_CheckSum(self, val, byt):
        sum = 0
        mask = 0xff  # 8 bits mask
        while (val):
            # print(sum, val)
            sum += val & mask  # only add the lower values
            val = val >> 8  # shift down
        sum += 0x9b
        # print(sum, byt)
        return (sum & 0xff) ^ byt  # if sum equal byt, this will be 0

    # waiting for a busy end, just for ADC1
    def ADS1263_WaitDRDY(self):
        for i in range(0, 400000, 1):
            if (config.digital_read(self.drdy_pin) == 0):
                break
        if (i >= 400000):
            print("Time Out ...\r\n")

    # Check chip ID, success is return 1
    def ADS1263_ReadChipID(self):
        id = self.ADS1263_ReadData(ADS1263_REG['REG_ID'])
        return id[0] >> 5

    def ADS1263_SetMode(self, Mode):
        ScanMode = Mode

    # The configuration parameters of ADC, gain and data rate
    def ADS1263_ConfigADC(self, gain, drate):
        MODE2 = 0x80
        MODE2 |= (gain << 4) | drate
        self.ADS1263_WriteReg(ADS1263_REG['REG_MODE2'], MODE2)
        if (self.ADS1263_ReadData(ADS1263_REG['REG_MODE2'])[0] == MODE2):
            print("REG_MODE2 success")
        else:
            print("REG_MODE2 unsuccess")

        REFMUX = 0x24
        self.ADS1263_WriteReg(ADS1263_REG['REG_REFMUX'], REFMUX)
        if (self.ADS1263_ReadData(ADS1263_REG['REG_REFMUX'])[0] == REFMUX):
            print("REG_REFMUX success")
        else:
            print("REG_REFMUX unsuccess")

        MODE0 = ADS1263_DELAY['ADS1263_DELAY_8d8ms']
        self.ADS1263_WriteReg(ADS1263_REG['REG_MODE0'], MODE0)
        if (self.ADS1263_ReadData(ADS1263_REG['REG_MODE0'])[0] == MODE0):
            print("REG_MODE0 success")
        else:
            print("REG_MODE0 unsuccess")

    # The configuration parameters of ADC2, gain and data rate
    def ADS1263_ConfigADC2(self, gain, drate):
        ADC2CFG = 0x20  # REF, 0x20:VAVDD and VAVSS, 0x00:+-2.5V
        ADC2CFG |= (drate << 6) | gain
        self.ADS1263_WriteReg(ADS1263_REG['REG_ADC2CFG'], ADC2CFG)
        if (self.ADS1263_ReadData(ADS1263_REG['REG_ADC2CFG'])[0] == ADC2CFG):
            print("REG_ADC2CFG success")
        else:
            print("REG_ADC2CFG unsuccess")

        MODE0 = ADS1263_DELAY['ADS1263_DELAY_8d8ms']
        self.ADS1263_WriteReg(ADS1263_REG['REG_MODE0'], MODE0)
        if (self.ADS1263_ReadData(ADS1263_REG['REG_MODE0'])[0] == MODE0):
            print("REG_MODE0 success")
        else:
            print("REG_MODE0 unsuccess")

    # Set ADC1 Measuring channel
    def ADS1263_SetChannal(self, Channal):
        if Channal > 10:
            return 0
        INPMUX = (Channal << 4) | 0x0a
        self.ADS1263_WriteReg(ADS1263_REG['REG_INPMUX'], INPMUX)
        if (self.ADS1263_ReadData(ADS1263_REG['REG_INPMUX'])[0] == INPMUX):
            # print("REG_INPMUX success")
            pass
        else:
            print("REG_INPMUX unsuccess")

    # Set ADC2 Measuring channel
    def ADS1263_SetChannal_ADC2(self, Channal):
        if Channal > 10:
            return 0
        INPMUX = (Channal << 4) | 0x0a
        self.ADS1263_WriteReg(ADS1263_REG['REG_ADC2MUX'], INPMUX)
        if (self.ADS1263_ReadData(ADS1263_REG['REG_ADC2MUX'])[0] == INPMUX):
            # print("REG_ADC2MUX success")
            pass
        else:
            print("REG_ADC2MUX unsuccess")

    # Set ADC1 Measuring differential channel
    def ADS1263_SetDiffChannal(self, Channal):
        if Channal == 0:
            INPMUX = (0 << 4) | 1  # DiffChannal	AIN0-AIN1
        elif Channal == 1:
            INPMUX = (2 << 4) | 3  # DiffChannal	AIN2-AIN3
        elif Channal == 2:
            INPMUX = (4 << 4) | 5  # DiffChannal	AIN4-AIN5
        elif Channal == 3:
            INPMUX = (6 << 4) | 7  # DiffChannal	AIN6-AIN7
        elif Channal == 4:
            INPMUX = (8 << 4) | 9  # DiffChannal	AIN8-AIN9
        self.ADS1263_WriteReg(ADS1263_REG['REG_INPMUX'], INPMUX)
        if (self.ADS1263_ReadData(ADS1263_REG['REG_INPMUX'])[0] == INPMUX):
            # print("REG_INPMUX success")
            pass
        else:
            print("REG_INPMUX unsuccess")

    # Set ADC2 Measuring differential channel
    def ADS1263_SetDiffChannal_ADC2(self, Channal):
        if Channal == 0:
            INPMUX = (0 << 4) | 1  # DiffChannal	AIN0-AIN1
        elif Channal == 1:
            INPMUX = (2 << 4) | 3  # DiffChannal	AIN2-AIN3
        elif Channal == 2:
            INPMUX = (4 << 4) | 5  # DiffChannal	AIN4-AIN5
        elif Channal == 3:
            INPMUX = (6 << 4) | 7  # DiffChannal	AIN6-AIN7
        elif Channal == 4:
            INPMUX = (8 << 4) | 9  # DiffChannal	AIN8-AIN9
        self.ADS1263_WriteReg(ADS1263_REG['REG_ADC2MUX'], INPMUX)
        if (self.ADS1263_ReadData(ADS1263_REG['REG_ADC2MUX'])[0] == INPMUX):
            # print("REG_ADC2MUX success")
            pass
        else:
            print("REG_ADC2MUX unsuccess")

    # Device initialization
    def ADS1263_init(self):
        if (config.module_init() != 0):
            return -1
        self.ADS1263_reset()
        id = self.ADS1263_ReadChipID()
        if id == 0x01:
            print("ID Read success  ")
        else:
            print("ID Read failed   ")
            return -1
        self.ADS1263_WriteCmd(ADS1263_CMD['CMD_STOP1'])
        self.ADS1263_WriteCmd(ADS1263_CMD['CMD_STOP2'])
        self.ADS1263_ConfigADC(ADS1263_GAIN['ADS1263_GAIN_1'], ADS1263_DRATE['ADS1263_20SPS'])
        self.ADS1263_ConfigADC2(ADS1263_ADC2_GAIN['ADS1263_ADC2_GAIN_1'], ADS1263_ADC2_DRATE['ADS1263_ADC2_10SPS'])
        return 0

    # Read ADC data
    def ADS1263_Read_ADC_Data(self):
        config.digital_write(self.cs_pin, GPIO.LOW)  # cs  0
        while (1):
            config.spi_writebyte([ADS1263_CMD['CMD_RDATA1']])
            config.delay_ms(10)
            if (config.spi_readbytes(1)[0] & 0x40 != 0):
                break
        buf = config.spi_readbytes(5)
        config.digital_write(self.cs_pin, GPIO.HIGH)  # cs 1
        read = (buf[0] << 24) & 0xff000000
        read |= (buf[1] << 16) & 0xff0000
        read |= (buf[2] << 8) & 0xff00
        read |= (buf[3]) & 0xff
        CRC = buf[4]
        # print(read, CRC)
        if (self.ADS1263_CheckSum(read, CRC) != 0):
            print("ADC1 data read error!")
        return read

    # Read ADC2 data
    def ADS1263_Read_ADC2_Data(self):
        read = 0
        config.digital_write(self.cs_pin, GPIO.LOW)  # cs  0
        while (1):
            config.spi_writebyte([ADS1263_CMD['CMD_RDATA2']])
            config.delay_ms(10)
            if (config.spi_readbytes(1)[0] & 0x80 != 0):
                break
        buf = config.spi_readbytes(5)
        config.digital_write(self.cs_pin, GPIO.HIGH)  # cs 1
        read |= (buf[0] << 16) & 0xff0000
        read |= (buf[1] << 8) & 0xff00
        read |= (buf[2]) & 0xff
        CRC = buf[4]
        if (self.ADS1263_CheckSum(read, CRC) != 0):
            print("ADC2 data read error!")
        return read

    # Read ADC1 specified channel data
    def ADS1263_GetChannalValue(self, Channel):
        if (ScanMode == 0):  # 0  Single-ended input  8 channel1 Differential input  4 channe
            if (Channel > 10):
                return 0
            self.ADS1263_SetChannal(Channel)
            config.delay_ms(2)
            self.ADS1263_WriteCmd(ADS1263_CMD['CMD_START1'])
            config.delay_ms(2)
            self.ADS1263_WaitDRDY()
            Value = self.ADS1263_Read_ADC_Data()
        else:
            if (Channel > 4):
                return 0
            self.ADS1263_SetDiffChannal(Channel)
            config.delay_ms(2)
            self.ADS1263_WriteCmd(ADS1263_CMD['CMD_START1'])
            config.delay_ms(2)
            self.ADS1263_WaitDRDY()
            Value = self.ADS1263_Read_ADC_Data()
        return Value

    # Read ADC2 specified channel data
    def ADS1263_GetChannalValue_ADC2(self, Channel):
        if (ScanMode == 0):  # 0  Single-ended input  8 channel1 Differential input  4 channe
            if (Channel > 10):
                return 0
            self.ADS1263_SetChannal_ADC2(Channel)
            config.delay_ms(2)
            self.ADS1263_WriteCmd(ADS1263_CMD['CMD_START2'])
            config.delay_ms(2)
            Value = self.ADS1263_Read_ADC2_Data()
        else:
            if (Channel > 4):
                return 0
            self.ADS1263_SetDiffChannal_ADC2(Channel)
            config.delay_ms(2)
            self.ADS1263_WriteCmd(ADS1263_CMD['CMD_START2'])
            config.delay_ms(2)
            Value = self.ADS1263_Read_ADC2_Data()
        return Value

    def ADS1263_GetAll(self):
        ADC_Value = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        for i in range(0, 10, 1):
            ADC_Value[i] = self.ADS1263_GetChannalValue(i)
            self.ADS1263_WriteCmd(ADS1263_CMD['CMD_STOP1'])
            config.delay_ms(20)
        print("--- Read ADC1 value success ---")
        return ADC_Value

    def ADS1263_GetAll_ADC2(self):
        ADC_Value = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        for i in range(0, 10, 1):
            ADC_Value[i] = self.ADS1263_GetChannalValue_ADC2(i)
            self.ADS1263_WriteCmd(ADS1263_CMD['CMD_STOP2'])
            config.delay_ms(20)
        print("--- Read ADC2 value success ---")
        return ADC_Value

    def ADS1263_RTD_Test(self):
        Delay = ADS1263_DELAY['ADS1263_DELAY_8d8ms']
        Gain = ADS1263_GAIN['ADS1263_GAIN_1']
        Drate = ADS1263_DRATE['ADS1263_20SPS']

        # MODE0 (CHOP OFF)
        MODE0 = Delay
        self.ADS1263_WriteReg(ADS1263_REG['REG_MODE0'], MODE0)
        config.delay_ms(1)

        # (IDACMUX) IDAC2 AINCOM,IDAC1 AIN3
        IDACMUX = (0x0a << 4) | 0x03
        self.ADS1263_WriteReg(ADS1263_REG['REG_IDACMUX'], IDACMUX)
        config.delay_ms(1)

        # ((IDACMAG)) IDAC2 = IDAC1 = 250uA
        IDACMAG = (0x03 << 4) | 0x03
        self.ADS1263_WriteReg(ADS1263_REG['REG_IDACMAG'], IDACMAG)
        config.delay_ms(1)

        MODE2 = (Gain << 4) | Drate
        self.ADS1263_WriteReg(ADS1263_REG['REG_MODE2'], MODE2)
        config.delay_ms(1)

        # INPMUX (AINP = AIN7, AINN = AIN6)
        INPMUX = (0x07 << 4) | 0x06
        self.ADS1263_WriteReg(ADS1263_REG['REG_INPMUX'], INPMUX)
        config.delay_ms(1)

        # REFMUX AIN4 AIN5
        REFMUX = (0x03 << 3) | 0x03
        self.ADS1263_WriteReg(ADS1263_REG['REG_REFMUX'], REFMUX)
        config.delay_ms(1)

        # Read one conversion
        self.ADS1263_WriteCmd(ADS1263_CMD['CMD_START1'])
        config.delay_ms(10)
        self.ADS1263_WaitDRDY()
        Value = self.ADS1263_Read_ADC_Data()
        self.ADS1263_WriteCmd(ADS1263_CMD['CMD_STOP1'])

        return Value

    def ADS1263_DAC_Test(self, isPositive, isOpen):
        Volt = ADS1263_DAC_VOLT['ADS1263_DAC_VLOT_3']

        if (isPositive):
            Reg = ADS1263_REG['REG_TDACP']  # IN6
        else:
            Reg = ADS1263_REG['REG_TDACN']  # IN7

        if (isOpen):
            Value = Volt | 0x80
        else:
            Value = 0x00

        self.ADS1263_WriteReg(Reg, Value)

    def ADS1263_Exit(self):
        config.module_exit()

### END OF FILE ###



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on July 22, 2021

Copyright Alpes Lasers SA, Neuchatel, Switzerland, 2021

@author: olgare
"""

from setuptools import setup

setup(
    setup_requires=['pbr'],
    pbr=True,
    test_suite = "phootonics_controller.tests"
)


import struct
from ..tools.definitions import ENDIAN, FRAME_BOUNDARY, ESCAPE_BYTE, ESCAPE_MASK, CommandID


def parse_error_flags(error_flag_data: int):
    r"""
    Parses error flag register values. Given error number, returns error string.

    Parameters
    ----------
    error_flag_data : int
        Error code hex number, will convert if string.

    Returns
    -------
    code: str
        Error code dictionary.
    """
    bits = format(error_flag_data[0], '#032b')[2::]
    error_result = {
        'Proxy Disconnected': bool(int(bits[29])),
        'Proxy Temperature Threshold Reached': bool(int(bits[28])),
        'Mirror Temperature Threshold Reached': bool(int(bits[27])),
        'Mirror EEPROM Invalid': bool(int(bits[26])),
        'Mirror Unstable': bool(int(bits[25])),
        'Linear Output Limit Reached': bool(int(bits[24])),
        'Linear Output Average Limit Reached': bool(int(bits[23])),
        'XY Input Trimmed': bool(int(bits[22])),
        'Proxy Was Disconnected': bool(int(bits[21])),
        'Proxy Temperature Threshold Was Reached': bool(int(bits[20])),
        'Mirror Temperature Threshold Was Reached': bool(int(bits[19])),
        'Linear Output Limit Was Reached': bool(int(bits[18])),
        'Linear Output Average Limit Was Reached': bool(int(bits[17])),
        'XY Input Was Trimmed': bool(int(bits[16])),
        'Reserved': bits[0:16]
    }

    return error_result


def parse_error(error_code: int):
    r"""
    Given error number, returns error string.

    Parameters
    ----------
    error_code : int
        Error code hex number, will convert if string.

    Examples
    --------
    Return error information from _board

    Returns
    -------
    code: str
        Error code string.

    >>> parse_error(0x0001)
    'COMMAND_LENGTH_MISMATCH'
    >>> parse_error('1')
    'COMMAND_LENGTH_MISMATCH'

    """

    error_codes = {
        0x0001:	'FRAME_INVALID',
        0x0002:	'COMMAND_UNKNOWN',
        0x0003:	'COMMAND_LENGTH_MISMATCH',
        0x0004:	'CRC_FAIL',
        0x0005:	'COMMAND_NOT_IMPLEMENTED_YET',

        0x0101:	'REGISTER_DOES_NOT_EXIST',
        0x0102:	'REGISTER_WRITE_UNSUCCESSFUL',
        0x0103:	'REGISTER_VALUE_OUT_OF_RANGE',
        0x0104:	'REGISTER_IS_NOT_VALID',
        0x0105: 'REGISTER_READ_UNSUCCESSFUL',
        0x0106: 'REGISTER_IS_READ_ONLY',
        0x0107: 'VECTOR_INDEX_OUT_OF_RANGE',
        0x0108: 'VECTOR_INDEX_NOT_ALIGNED',
        0x0109: 'REGISTER_IS_LOCKED',
        0x010a: 'REGISTER_IS_NOT_READABLE',

        0x0201:	'SYSTEM_DOES_NOT_EXIST',
        0x0202:	'SYSTEM_NOT_A_VALID_SELECTION',
        0x0203:	'CHAIN_CONSTRUCTION_ERROR',

        0x0701:	'ERROR_FLAG_NO_SNAPSHOT_VALUES_AVAILABLE_TO_LOAD',
        0x0702: 'ERROR_FLAG_CANNOT_SAVE_SNAPSHOT',
        0x0703: 'ERROR_FLAG_CORRUPTED_SNAPSHOT',
        0x0704: 'ERROR_FLAG_SNAPSHOT_INDEX_OUT_OF_RANGE',
        0x0705: 'ERROR_FLAG_NOT_ENOUGH_SPACE_FOR_SNAPSHOT',
        0x0f0f: 'COMMUNICATION_ERROR',
        0xffff: 'WRONG_BOARD_MODE_FOR_COMMAND'
        }
    code = 'UNKNOWN_ERROR_CODE'
    try:
        code = error_codes[int(error_code)]
    finally:
        return code


def encode(command_id, register_id, data=None):
    r"""
    Given command_id, register_id and payload, will determine proper bytestring to transmit.
    Uses the following format:

    +----------+---------------------+---------+-----------+------------+---------------+----------+
    | Frame    | Reserved            | Command | Data size | Data       | CRC           | Frame    |
    | Boundary | slave address: 0x00 |         |           | payload    | if applicable | Boundary |
    +==========+=====================+=========+===========+============+===============+==========+
    | 1 byte   | 1 byte              | 1 byte  | 1 byte    | 0-50 bytes | 2 bytes       | 1 byte   |
    +----------+---------------------+---------+-----------+------------+---------------+----------+

    Parameters
    ----------
    command_id
        Identifier corresponding to command
    register_id
        Identifier corresponding to register. See register['id'] field.
    data
        Float or int or bytes, depending on the type of register

    Raises
    ------
    none

    Returns
    -------
    b_message: bytearray
        Stuffed and escaped bytestring, following protocol above.

    Notes
    -----
    Given a FRAME_BOUNDARY of 0x7e and ESCAPE_BYTE of 0x7d, every appearance of the bytes 0x7e or 0x7d in the data
    (so in a role other than the delimiter or the escape byte)  is replaced by the sequence 0x7d 0x5e or 0x7d 0x5d,
    respectively. This follows ESCAPE_MASK = 1 << 5, or 0x20, thus the byte is XOR'd with the mask and prepended by 0x7d
    """
    b_header = b''
    b_message = b''
    b_delimiter = struct.pack(">B", 0x7E)
    b_slave_address = struct.pack(">B", 0x00)
    b_crc = struct.pack(">H", 0x00)
    b_cocommand_id = struct.pack(">B", command_id)
    if isinstance(register_id, list):  # Used for multiple values commands
        b_register_id = struct.pack(">H", len(register_id))
        for x in register_id:
            b_register_id += struct.pack(">H", x)
        if data is None:  # get values command
            size = 2 + 2 * len(register_id)
            b_header = b_slave_address + b_cocommand_id + struct.pack(">B", size)
            b_message = b_delimiter + b_header + b_register_id + b_crc + b_delimiter
        else:  # set values command
            size = 2 + 2 * len(register_id) + 4 * len(data)
            b_header = b_slave_address + b_cocommand_id + struct.pack(">B", size)
            b_data = b''
            for v in data:
                if isinstance(v, float):
                    b_data += struct.pack(">f", v)
                elif isinstance(v, int):
                    b_data += struct.pack(">i", v)
            b_message = b_delimiter + b_header + b_register_id + b_data + b_crc + b_delimiter
    else:  # Used for set get value and vector
        b_register_id = struct.pack(">H", register_id)
        if data is not None:
            if isinstance(data, float):  # Used for set value, when value is float
                b_data = struct.pack(">f", data)
                b_header = b_slave_address + b_cocommand_id + struct.pack(">B", 0x06)
            elif isinstance(data, int):  # Used for set value, when value is float
                b_data = struct.pack(">i", data)
                b_header = b_slave_address + b_cocommand_id + struct.pack(">B", 0x06)
            elif isinstance(data, list):  # Used for get Vector
                if len(data) == 2:
                    b_data = struct.pack(">H", data[0]) + struct.pack(">B", data[1])  # Pack vector index and length
                    b_header = b_slave_address + b_cocommand_id + struct.pack(">B", 0x05)
            b_message = b_delimiter + b_header + b_register_id + b_data + b_crc + b_delimiter
        else:
            if command_id == 0x04 or command_id == 0x05:
                b_header = b_slave_address + b_cocommand_id + struct.pack(">B", 0x01)  # Payload contains 1 byte config ID
                b_message = b_delimiter + b_header + struct.pack(">B", register_id) + b_crc + b_delimiter
            else:
                b_header = b_slave_address + b_cocommand_id + struct.pack(">B", 0x02)  # Payload contains only registers ID
                b_message = b_delimiter + b_header + b_register_id + b_crc + b_delimiter
    b_message = b_message.replace(b'}', b'}]')
    b_message = b_message[1:-1].replace(b'~', b'}^')
    b_message = b'~' + b_message + b'~'

    return b_message


def encode_vector(register, index, vector):
    """
    Encodes the given register vector to bytes.

    Parameters
    ----------
    register : dict or int
        Register dictionary corresponding to desired registers.{'id': registers number, 'type': data type, ...}
        Further, a raw registers id can be passed, in which case no validation will occur (COULD CAUSE ERRORS).
    index : int
        The location to store the given vector.
    value : varies, can also be list of values
        The type and range of value is dependent on the registers being addressed, see 'type' element above.
    cmd_delay : float
        How long to wait for response in secs. By default is set to registers.CMD_DELAY


    Returns
    -------
    str
        byte string to transmit.

    """
    cmd_id = CommandID.SET_VECTOR
    index = index * 4
    bstr = bytes([])

    if type(register) is int:
        # generic packing, vals must already be in correct format for transmission
        vals = struct.pack(ENDIAN + "f" * len(vector), *vector)
        data = struct.pack(ENDIAN + "HHB", register, index, len(vals)) + vals
    else:
        if register['type'] is int:
            # common, int32 values
            vals = struct.pack(ENDIAN + "i" * len(vector), *vector)
            data = struct.pack(ENDIAN + "HHB", register['id'], index, len(vals)) + vals
        elif register['type'] is float:
            # common, float32 values
            vals = struct.pack(ENDIAN + "f" * len(vector), *vector)
            data = struct.pack(ENDIAN + "HHB", register['id'], index, len(vals)) + vals
        elif register['type'] is bytes:
            # less common, generally for eeprom data
            vals = bytes(list(vector))
            data = struct.pack(ENDIAN + "HHB", register['id'], index // 4, len(vals)) + vals
        elif register['type'] is str:
            # uncommon, char values
            vals = struct.pack(ENDIAN + "c" * len(vector), *vector)
            data = struct.pack(ENDIAN + "HHB", register['id'], index // 4, len(vals)) + vals
        else:
            # generic stuffing due to unknown register['type'] value
            bstr = bytes([0, cmd_id, len(vector)] + list(vector))
            bstr = bstr + bytes((0, 0))

            def stuff(unstuffed):
                result = [FRAME_BOUNDARY]
                for b in unstuffed:
                    if b in (ESCAPE_BYTE, FRAME_BOUNDARY):
                        result.append(ESCAPE_BYTE)
                        result.append(b ^ ESCAPE_MASK)
                    else:
                        result.append(b)
                result.append(FRAME_BOUNDARY)
                return bytes(result)
            return stuff(bstr)

    # build packet frame and stuff bytes
    bstr = bytes([FRAME_BOUNDARY, 0, cmd_id, len(data)] + list(data) + [0, 0, FRAME_BOUNDARY])
    bstr = bstr.replace(bytes([ESCAPE_BYTE]), bytes([ESCAPE_BYTE, ESCAPE_BYTE ^ ESCAPE_MASK]))
    bstr = bstr[1:-1].replace(bytes([FRAME_BOUNDARY]), bytes([ESCAPE_BYTE, FRAME_BOUNDARY ^ ESCAPE_MASK]))
    bstr = b'~' + bstr + b'~'
    return bstr


def decode(data: bytes):
    r"""
    Given bytestring, will unpack/destuff values, following stuffing method described in Notes.

    Parameters
    ----------
    command_id
        Identifier corresponding to command
    register_id
        Identifier corresponding to register. See register['id'] field.

    Raises
    ------
    none

    Returns
    -------
    result: bytearray
        Destuffed bytestring, following protocol above.

    Notes
    -----
    Given a FRAME_BOUNDARY of 0x7e and ESCAPE_BYTE of 0x7d, every appearance of the bytes 0x7e or 0x7d in the data
    (so in a role other than the delimiter or the escape byte)  is replaced by the sequence 0x7d 0x5e or 0x7d 0x5d,
    respectively. This follows ESCAPE_MASK = 1 << 5, or 0x20, thus the byte is XOR'd with the mask and prepended by 0x7d
    """
    result = []
    try:
        if len(data) < 2 or data[0] != FRAME_BOUNDARY or data[-1] != FRAME_BOUNDARY:
            print('Byte De-stuffing Cannot Be Performed. [data={}]'.format(data))
            return None

        i = 1
        while i < (len(data) - 1):
            if data[i] == ESCAPE_BYTE:
                result.append(data[i + 1] ^ ESCAPE_MASK)
                i += 2
            else:
                result.append(data[i])
                i += 1
    except TypeError or KeyError:
        print('Byte De-stuffing Cannot Be Performed. [data={}]'.format(data))

    if FRAME_BOUNDARY in data[1:-1]:
        print("ERROR: Duplicate Frame Returned. System Reboot Required.")
        return None
    else:
        return bytes(result)



from external_cavity.controllers.driver_utils.MR_E_2_Python_SDK.optoKummenberg.tools.systems_registers_tools import *


def list_systems():
    """
    Inspect generic_registers for all imported and defined Systems.
    :return: list of System names
    """
    all_systems = inspect.getmembers(sys.modules['optoMDC.registers.xpr4_registers'], inspect.isclass)
    return [system[0] for system in all_systems if hasattr(system[1](0), '_is_a_system')]


def get_register(system_name: str, register_name, channel: int = 0):
    """
    Given a System name and Register name, will return the associated register_id integer
    :param system_name: str
    :param register_name: str
    :param channel: int
    :return: int register_id
    """
    registers = inspect.getmembers(sys.modules['optoMDC.registers.xpr4_registers'], inspect.isclass)
    register_system = dict(registers)[system_name](channel)
    register_list = inspect.getmembers(register_system)
    register_id = dict(register_list)[register_name]
    return register_id


import logging
import struct
import time
import serial
from ..tools.parsing_tools import parse_error, encode, decode
from ..tools.definitions import (CMD_DELAY, ENDIAN, FRAME_BOUNDARY,
                                 SimpleResponses, commandID, commandName,
                                 WaveformShape, UnitType)
from ..tools.systems_registers_tools import process_registers
import numpy as np


logger = logging.getLogger(__name__)

def issue_command(board, command_bytes, cmd_delay: float = None):
    r"""
    Issues given command_bytes over _board connection. Will attempt 3 times.

    This method unlocks the comm_lock, sends the desired data, receives a response, then reapplies the lock.

    Parameters
    ----------
    board : Board
        The board to issue a command over.
    command_bytes : bytes or int or str
        The byte array to send over the connection.

    Returns
    -------
    str
        Firmware response, contains acknowledgement or parse_error information. In DEBUG mode this returns the internal state.

    Raises
    ------
    none

    See Also
    --------
    Command bytes must be properly encoded, see registers encode function.

    """
    board.Connection._comm_lock = False
    number_of_attempts = 0
    response = b''

    if cmd_delay is None:
        cmd_delay = CMD_DELAY

    while response == b'' and number_of_attempts < 3:
        try:
            # clear output buffer
            temp = board.Connection.verbose
            board.Connection.verbose = False
            junk = board.Connection.receive()
            board.Connection.verbose = temp
            if junk != b'':
                response = junk
                break
            if board.Connection.verbose and junk != b'':
                logger.debug('Output Buffer Cleared. [data={}]'.format(junk))

            # call and respond
            board.Connection.send(command_bytes)
            time.sleep(cmd_delay)
            if not hasattr(board, '_simple'):
                reponse = board.Connection.receive()
            else:
                if board._simple:
                    response = board.Connection.receive(terminator='\n')
                elif not board._simple:
                    response = board.Connection.receive(terminator=bytes([FRAME_BOUNDARY]))
            board.Connection._serial_conn.flush()
            number_of_attempts += 1
            break

        # exception handling
        except serial.serialutil.SerialException:
            err_args = serial.serialutil.SerialException.args
            logger.debug('Unable to Communicate.')
            logger.debug('Retrying...')
            response = err_args
            logger.debug(err_args)
            number_of_attempts += 1
        except AttributeError:
            err_args = serial.serialutil.SerialException.args
            logger.debug('Serial Connection does not exist. Is the port ID valid?')
            logger.debug('Retrying...')
            response = err_args
            number_of_attempts += 1

    # Check connection after 3 failed attempts. Try to reconnect
    if number_of_attempts >= 3:
        attempt_reconnect(board)

    board.Connection._comm_lock = True

    return response


def attempt_reconnect(board):
    r"""
    Will attempt to reconnect to board. Will attempt 3 times.

    Parameters
    ----------
    board : Board
        The board to attempt reconnect.

    Returns
    -------
    none

    Raises
    ------
    Upon failure, will dump serial information.

    """
    try:
        logger.debug('Connection Failed. Resetting Connection and Synchronizing to Board...')
        board.disconnect()
        time.sleep(1)
        board.Connection._serial_conn.open()
    except:
        err_args = serial.serialutil.SerialException.args
        logger.debug('No Connection. Unable to Reconnect.')
        logger.debug('Argument Dump: {}'.format(err_args))
        logger.debug('Serial Connection: {}'.format(board.Connection._serial_conn))
        logger.debug('Serial Settings: {}'.format(board.Connection._settings))
        raise Exception('cannot re-connect VCM board after 3 failed attempts')


def process_simple_response(board, process_name: str, received_response: bytes):
    r"""
    Processes firmware simple response to determine if successful.

    Parameters
    ----------
    board : Board
        The object sending/receiving.
    process_name : str
        The name of the attempted process, for output statement.
    received_response : bytes
        The response received over the communication channel.

    Returns
    -------
    processed_response or error_response
        Decoded firmware response

    Raises
    ------
    none

    """
    if received_response == b'':
        return 'Empty Response...'

    if board._simple:
        if received_response == SimpleResponses['PROCESSED_WITHOUT_ERROR']:
            if board.verbose:
                logger.debug('{} Successful.'.format(process_name))
            processed_response = received_response
        elif received_response in SimpleResponses.values():
            if board.verbose:
                logger.debug('{} NOT Successful.'.format(process_name))
            processed_response = received_response
        else:
            err_args = parse_error(received_response)
            logger.debug('{} NOT Successful: {}.'.format(process_name, err_args))
            processed_response = err_args
    else:
        processed_response = parse_error(0xffff)
        logger.debug('{} NOT Successful: {}'.format(process_name, processed_response))

    return processed_response


def process_response(command_id_sent: int, register, value, response, verbose=False):
    r"""
    Processes firmware response.

    Parameters
    ----------
    command_id_sent : int
        The attempted command, see: commandID dictionary for potential values.
    register : dict or list of dicts or int or list of ints
        Register dictionary corresponding to desired registers.{'id': registers number, 'type': data type, ...}
    value
        The value or values that were originally destined to be transmitted.
    response : bytes or int or str
        Firmware response.
    verbose : bool = False
        Output verbosity.
    simulated : bool = False
        Whether connection over port is simulated, thus response is simulated.

    Returns
    -------
    slave_id, command_id, response_data, crc
        Decoded firmware response

    Raises
    ------
    none

    """
    if type(response) is not bytes or response == b'':
        return 0, 0, response, 0

    processed_response = decode(response)
    if processed_response is None:
        # comm parse_error
        logger.debug('ERROR: Communication Error.')
        return 0, 0, [0x0, 0xf, 0x0, 0xf], 0, True
    slave_id = processed_response[0]
    command_id  = processed_response[1]

    response_data = []
    if command_id == commandID['GET_VALUE']:
        # get single
        data_size = processed_response[2]
        raw_data = processed_response[3:data_size + 3]
        if type(register) is int:
            # return raw bytes, since only register id given
            response_data = raw_data
        elif register['type'] is int:
            response_data.append(struct.unpack(ENDIAN+'i', raw_data)[0])
        elif register['type'] is float:
            response_data.append(struct.unpack(ENDIAN+'f', raw_data)[0])
        elif register['type'] is bool:
            response_data.append(struct.unpack(ENDIAN+'i', raw_data)[0])

    if command_id == commandID['GET_MULTIPLE']:
        # get multiple
        data_size = processed_response[4]
        for i in range(data_size):
            start_byte = 1 + 4*(1+i)
            end_byte   = 5 + 4*(1+i)
            data_bytes = processed_response[start_byte:end_byte]
            if register[i]['type'] is int:
                data_bytes = struct.unpack(ENDIAN+'i', data_bytes)[0]
            elif register[i]['type'] is float:
                data_bytes = struct.unpack(ENDIAN+'f', data_bytes)[0]
            elif register[i]['type'] is bool:
                data_bytes = struct.unpack(ENDIAN+'i', data_bytes)[0]
            response_data.append(data_bytes)

    if command_id_sent == commandID['GET_VECTOR']:
        data_size = processed_response[2]
        if register['type'] in [str, bytes]:
            data_bytes = processed_response[3:data_size+3]
            response_data.append(data_bytes)
        elif data_size > 0:
            for i in range(data_size//4):
                start_byte = 3 + 4 * i
                end_byte = 7 + 4 * i
                data_bytes = processed_response[start_byte:end_byte]
                if register['type'] is int:
                    data_bytes = struct.unpack(ENDIAN + 'i', data_bytes)[0]
                elif register['type'] is float:
                    data_bytes = struct.unpack(ENDIAN + 'f', data_bytes)[0]
                response_data.append(data_bytes)

    if command_id_sent | 0x80 == command_id:
        # parse_error response!
        data_size = processed_response[2]
        response_data = processed_response[3:data_size + 3]
        # response_data = struct.unpack(ENDIAN+'i', response_data)

    response_data = list(response_data)
    crc = [processed_response[-2], processed_response[-1]]

    command_name = commandName[command_id_sent]
    err = True

    # success
    if command_id_sent == command_id:
        err = False
        if isinstance(register, list):
            register_id = [reg['id'] for reg in register]
            register_names = [hex(reg_id) for reg_id in register_id]
            if verbose:
                logger.debug('Command: [{}] Registers: [{}]. Values: [{}].'
                      .format(command_name, register_names, response_data))
        elif isinstance(register, dict):
            register_id = register['id']
            if verbose:
                logger.debug('Command: [{}]. Register: [{}]. Value: {}.'.format(command_name, hex(register_id), response_data))

    # not success
    elif command_id_sent != command_id:
        logger.debug('ERROR: Command ID Mismatch.')
        logger.debug('ID Expected: {}. ID Received: {}'.format(hex(command_id_sent), hex(command_id)))
        if command_id_sent | 0x80 == command_id:
            code = response_data[0] << 24 | \
                   response_data[1] << 16 | \
                   response_data[2] << 8 | \
                   response_data[3] << 0
            err_code = parse_error(code)
            logger.debug('This Corresponds to Error Code: {}>>{}'.format(code, err_code))
        else:
            logger.debug('This Corresponds to an Unknown Error Code.')
        logger.debug('Received Data Dump: {}'.format(response))

    return slave_id, command_id, response_data, crc, err


def get_cmd_reg_val(command, register: dict or list or int, value=None):
    r"""
    Given register(s) and value(s) will return command_id, register_id, value.

    Parameters
    ----------
    command : string or int related to command type 'SET_VALUE', 'GET_VALUE', etc with no determination of MULTIPLE
    register : dict or list of dicts or int or list of ints
        Register dictionary corresponding to desired registers.{'id': registers number, 'type': data type, ...}
        Further, a raw registers id can be passed, in which case no validation will occur (COULD CAUSE ERRORS).
    value : varies, can also be list of values
        The type and range of value is dependent on the registers being addressed, see 'type' element above.

    Returns
    -------
    cmd_id
        Hex value corresponding to proper command (1 Byte).
    reg_id
        Hex value corresponding to proper register (2 Bytes).
    value
        Validated value for the given register.

    Raises
    ------
    none

    See Also
    --------
    Register IDs, value type and range can be found in the registers module.

    Examples
    --------
    In this example, set value command for static input OF on channel 1 to 0.1 is expected.

    >>> from optoKummenberg.registers.generic_registers import StaticInput
    >>> res=get_cmd_register_value(0x10, 0x5100, 0.1)
    >>> print(encode(*res))
    b'~\x00\x10\x06Q\x00=\xcc\xcc\xcd\x00\x00~'
    >>> res=get_cmd_register_value(commandID['SET_VALUE'], StaticInput(1).current, 0.1)
    >>> print(encode(*res))
    b'~\x00\x10\x06Q\x00=\xcc\xcc\xcd\x00\x00~'
    >>> res=get_cmd_register_value('SET_VALUE', StaticInput(1).current, 0.1)
    >>> print(encode(*res))
    b'~\x00\x10\x06Q\x00=\xcc\xcc\xcd\x00\x00~'

    You can see these results are the same, thus any of these methods are valid.
    """
    # Convert command to hex if needed
    if type(command) is str:
        try:
            cmd_id = commandID[command]
        except KeyError:
            logger.debug("Unknown Command: {}".format(command))
            logger.debug("Valid Commands: {}".format(list(commandID.keys())))
            return None, None, None
    else:
        cmd_id = command

    # Convert command to multiple if needed
    if type(register) is list:
        if cmd_id == commandID['SET_VALUE']:
            cmd_id = commandID['SET_MULTIPLE']
        elif cmd_id == commandID['GET_VALUE']:
            cmd_id = commandID['GET_MULTIPLE']

    # Use register_tables if needed
    if type(value) in [UnitType, WaveformShape]:
        value = value.value

    # int32 and float64 are not captured by int and float, so we first convert to them
    if type(value) is np.int32:
        value = int(value)
    if type(value) is np.float64:
        value = float(value)

    # for setting manager value to a stage for a given channel & system
    if type(value) not in (int, list, bool, float, type(None)):
        if type(value) is str:
            found = False
            for key, field in list(register['range'].items()):
                if found:
                    break
                if field.lower().upper() == value.lower().upper():
                    found = True
                    break
            if found:
                value = key
        else:
            try:
                value = value.sys_id
            except AttributeError:
                logger.debug("Invalid System")
                return SimpleResponses['UNRECOGNIZED_MESSAGE']

    # set multiple with one value input
    if value is not None:
        if type(value) in (int, float, type(None)) and type(register) is list and len(register) > 1:
            value_list = []
            for i in range(len(register)):
                try:
                    if type(register[i]['type']) is int:
                        value_list.append(int(value))
                    elif type(register[i]['type']) is float:
                        value_list.append(float(value))
                    else:
                        value_list.append(int(value))
                except:
                    value_list.append(float(value))
            value = value_list

    # # handle various registers input types # #
    if type(register) in (int, list):
        # raw, raws or dicts
        if type(register) is list:
            # raws or dicts
            if all(isinstance(x, int) for x in register):
                # raws
                cmd_id, register_id = (cmd_id, register) if len(register) <= 8 else (None, None)
            else:
                # dicts
                if value is not None:
                    for i in range(len(value)):
                        if type(value[i]) is str:
                            found = False
                            for key, field in list(register[i]['range'].items()):
                                if found:
                                    break
                                if field.lower().upper() == value[i].lower().upper():
                                    found = True
                                    break
                            if found:
                                value[i] = key
                        elif type(value[i]) in [UnitType, WaveformShape]:
                            value[i] = value[i].value

                register_id = process_registers(register, value)
        else:
            # raw
            register_id = register
    else:
        # dict
        register_id = process_registers(register, value)
        if register_id is None:
            return None, None, None

    return (cmd_id, register_id, value) if register_id else (None, None, None)


def _set_internal(register: dict, response_data, value):
    """
    Currently unused.
    Intended for setting register['value'], such that state could be stored in a log file.
    :param register:
    :param response_data:
    :param value:
    :return:
    """
    if isinstance(value, list):
        value = value[0]
    while isinstance(response_data, list):
        while len(response_data) > 1:
            response_data = response_data.pop()
        if len(response_data) == 1:
            response_data = response_data[0]
        elif len(response_data) == 0:
            break

    if (value is None) and (response_data is None):
        register['value'] = None
    elif value == [] or value is None:
        register['value'] = response_data
    else:
        register['value'] = value


class System:
    _is_a_system = False

    def __init__(self, channel: int or None = None, board=None):
        self.name = self.__class__.__name__
        self._channel = channel
        self._is_a_system = True
        if channel is None:
            self._is_a_system = False
        self._board = board
        self.register_list = self._get_register_list()

    # TODO: this would run faster without lookup loop, just send hex to board.set/get
    def set_register(self, register_name_or_id: str or int, value,
                     set_internal: bool = False, cmd_delay: float = None):
        if self._board is not None:
            if isinstance(register_name_or_id, int):
                reg_id = register_name_or_id
                register_name = [item[0] for item in self.register_list if item[1]['id'] & 0x0f == reg_id][0]
            else:
                register_name = register_name_or_id
            reg = self.__dict__[register_name]
            if reg['type'] is float:
                value = float(value)
            response = self._board.set_value(reg, value, cmd_delay=cmd_delay)
            if isinstance(value, list):
                value = value[0]
            if set_internal:
                self.__dict__[register_name]['value'] = value
            return response
        return None

    def get_register(self, register_name_or_id: str or int):
        if self._board is not None:
            if isinstance(register_name_or_id, int):
                reg_id = register_name_or_id
                register_name = [item[0] for item in self.register_list if item[1]['id'] & 0x0f == reg_id][0]
            else:
                register_name = register_name_or_id
            response = self._board.get_value(self.__dict__[register_name])

            return response
        return None

    def get_register_names(self):
        return [item for item in self.__dict__ if type(self.__dict__[item]) is dict]

    def _get_register_list(self):
        return [[item, self.__dict__[item]] for item in self.__dict__ if type(self.__dict__[item]) is dict]


class InputStage(System):
    _is_a_system = False

    def __init__(self, channel: int or None = 0, board=None):
        System.__init__(self, channel, board)
        self.name = self.__class__.__name__

    def SetAsInput(self):
        ch = self._board.channel[self._channel]
        ch.Manager.set_register('input', self)


class ControlStage(System):
    _is_a_system = False

    def __init__(self, channel: int or None = 0, board=None):
        System.__init__(self, channel, board)
        self.name = self.__class__.__name__

    def SetAsControlMode(self):
        ch = self._board.channel[self._channel]
        ch.Manager.set_register('control', self)


from .ClassAbstracts import System
from ..tools.parsing_tools import parse_error_flags
from ..tools.definitions import CHUNK_SIZE, UnitType
from ..tools.systems_registers_tools import get_registers
from warnings import warn


class Manager(System):
    r"""
The Signal Flow Channel Manager is managing the active systems in the signal flow chain.
System ID for channel X: 0x40

+---------------------+---------------+-------------+----------+--------------+---------------+--------------------+
| Register Name       | Id            | Type        | Unit     | Range        | Default       | Comment            |
+=====================+===============+=============+==========+==============+===============+====================+
| input               | 5xCHANNEL + 0 | uint 32-bit | SystemID | 0x48 to 0x7f |               |                    |
+---------------------+---------------+-------------+----------+--------------+---------------+--------------------+
| input_conditioning  | 5xCHANNEL + 1 | uint 32-bit | SystemID | 0x48 to 0x7f |               |                    |
+---------------------+---------------+-------------+----------+--------------+---------------+--------------------+
| control             | 5xCHANNEL + 2 | uint 32-bit | SystemID | 0x48 to 0x7f |               |                    |
+---------------------+---------------+-------------+----------+--------------+---------------+--------------------+
| output_conditioning | 5xCHANNEL + 3 | uint 32-bit | SystemID | 0x48 to 0x7f |               |                    |
+---------------------+---------------+-------------+----------+--------------+---------------+--------------------+
| output              | 5*CHANNEL + 4 | uint 32-bit | SystemID | 0x48 to 0x7f |               |                    |
+---------------------+---------------+-------------+----------+--------------+---------------+--------------------+

    """

    @staticmethod
    def help():
        print(Manager.__doc__)

    def __init__(self, channel: int = 0, board=None):
        self.sys_id = 0x40
        self._readonly = False

        # TODO: I could calculate valid range here from channel using bit math.
        self.input = {'id': self.sys_id << 8 | channel * 5 | 0x00,
                      'type': int,
                      'unit': 'SystemID',
                      'range': [0x48, 0x7f],
                      'default': None,
                      'value': None}
        self.input_conditioning = {'name': 'input_conditioning',
                                   'id': self.sys_id << 8 | channel * 5 | 0x01,
                                   'type': int,
                                   'unit': 'SystemID',
                                   'range': [0x48, 0x7f],
                                   'default': None,
                                   'value': None}
        self.control = {'name': 'control',
                        'id': self.sys_id << 8 | channel * 5 | 0x02,
                        'type': int,
                        'unit': None,
                        'range': None,
                        'default': None,
                        'value': None}
        self.output_conditioning = {'name': 'output_conditioning',
                                    'id': self.sys_id << 8 | channel * 5 | 0x03,
                                    'type': int,
                                    'unit': 'SystemID',
                                    'range': [0x48, 0x7f],
                                    'default': None,
                                    'value': None}
        self.output = {'name': 'output',
                       'id': self.sys_id << 8 | channel * 5 | 0x04,
                       'type': int,
                       'unit': 'SystemID',
                       'range': [0x48, 0xef],
                       'default': None,
                       'value': None}

        System.__init__(self, channel, board)
        self.name = self.__class__.__name__

    def CheckSignalFlow(self):
        # pull input, input_conditioning, control, output_conditioning, and output registers
        input_id = self.get_register('input')[0]
        icond_id = self.get_register('input_conditioning')[0]
        cntrl_id = self.get_register('control')[0]
        ocond_id = self.get_register('output_conditioning')[0]
        outpt_id = self.get_register('output')[0]
        # get input units
        if input_id in range(0x50, 0x60):
            # static input has no units register
            input_units = 'Static'
        else:
            inpt_units_register = get_registers(self._board.systems[input_id]['name'])['unit']
            input_units = UnitType(self._board.get_value(inpt_units_register)[0]).name

        # find corresponding system for input and mode for control
        flow_states = {'SignalFlowManager': {'name': self.name,
                                             'channel': self._channel},
                       'InputStage': {'name': self._board.systems[input_id]['name'],
                                      'channel': input_id & 0x0F % 8,
                                      'unit': input_units},
                       'InputConditioning': {'name': self._board.systems[icond_id]['name'],
                                             'channel': icond_id & 0x0F % 8},
                       'ControlStage': {'name': self._board.systems[cntrl_id]['name'],
                                        'channel': cntrl_id & 0x0F % 8},
                       'OutputConditioning': {'name': self._board.systems[ocond_id]['name'],
                                              'channel': ocond_id & 0x0F % 8},
                       'OutputStage': {'name': self._board.systems[outpt_id]['name'],
                                       'channel': outpt_id & 0x0F % 8},
                       }
        # warning if unit mismatch 'except static input'
        if not flow_states['InputStage']['unit'].lower() in flow_states['ControlStage']['name'].lower() \
                and input_id not in range(0x50, 0x60):
            warn("Input Units: {} Do Not Match Control Mode: {}".format(flow_states['InputStage']['unit'],
                                                                        flow_states['ControlStage']['name']),
                 RuntimeWarning)
        # warning if channel mismatch
        used_channels = [flow_states[item]['channel'] for item in flow_states]
        if not all([item == used_channels[0] for item in used_channels]):
            warn("Not all stages are set to the same channel.", RuntimeWarning)
        return flow_states


class Status(System):
    r"""
Device Functionality - Firmware Status
System ID: 0x10

+---------------------+------+-------------+---------+---------------------------------------+
| Register Name       | Id   | Type        | Default | Comment                               |
+=====================+======+=============+=========+=======================================+
| firmware_id         | 0x00 | float 32-bit|         |                                       |
+---------------------+------+-------------+---------+---------------------------------------+
| firmware_branch     | 0x01 | float 32-bit|         |                                       |
+---------------------+------+-------------+---------+---------------------------------------+
| fw_type             | 0x02 | float 32-bit|         |                                       |
+---------------------+------+-------------+---------+---------------------------------------+
| fw_version_major    | 0x03 | float 32-bit|         |                                       |
+---------------------+------+-------------+---------+---------------------------------------+
| fw_version_minor    | 0x04 | float 32-bit|         |                                       |
+---------------------+------+-------------+---------+---------------------------------------+
| fw_version_build    | 0x05 | float 32-bit|         |                                       |
+---------------------+------+-------------+---------+---------------------------------------+
| fw_version_revision | 0x06 | float 32-bit|         |                                       |
+---------------------+------+-------------+---------+---------------------------------------+
| error_flag_register | 0x07 | float 32-bit| 0       | See: Status Register Values Below     |
+---------------------+------+-------------+---------+---------------------------------------+
| proxy_fpga_version  | 0x08 | float 32-bit|         |                                       |
+---------------------+------+-------------+---------+---------------------------------------+
| cpu_fpga_version    | 0x09 | float 32-bit|         |                                       |
+---------------------+------+-------------+---------+---------------------------------------+

Status Register Values:

====== =================================================================
Bit#   Message
====== =================================================================
0      Proxy not connected
1      Proxy temperature threshold is reached. See TemperatureManager
2      Mirror temperature threshold is reached See TemperatureManager
3      Mirror EEPROM not valid
4      Mirror not stable. See stability criterion in OpticalFeedback
5      Linear output limit is reached. See LinearOutput
6      Linear output average limit is reached. See LinearOutput
7      XY input is trimmed. See InputConditioning
8      Proxy was disconnected
9      Proxy temperature threshold was reached
10     Mirror temperature threshold was reached
11     Linear output limit was reached
12     Linear output average limit was reached
13     XY input was trimmed
14..31 Reserved
====== =================================================================
    Write:
    Writing to this register resets all history error flags 8, 9, 10, 11, 12, 13
    """

    @staticmethod
    def help():
        print(Status.__doc__)

    _is_a_system = False

    def __init__(self, board=None):

        self.sys_id = 0x10

        self.firmware_id = {'id': self.sys_id << 8 | 0x00,
                            'type': int,
                            'unit': None,
                            'range': None,
                            'default': None,
                            'value': None}
        self.fw_branch = {'id': self.sys_id << 8 | 0x01,
                          'type': int,
                          'unit': None,
                          'range': None,
                          'default': None,
                          'value': None}
        self.fw_type = {'id': self.sys_id << 8 | 0x02,
                        'type': int,
                        'unit': None,
                        'range': None,
                        'default': None,
                        'value': None}
        self.fw_version_major = {'id': self.sys_id << 8 | 0x03,
                                 'type': int,
                                 'unit': None,
                                 'range': None,
                                 'default': None,
                                 'value': None}
        self.fw_version_minor = {'id': self.sys_id << 8 | 0x04,
                                 'type': int,
                                 'unit': None,
                                 'range': None,
                                 'default': None,
                                 'value': None}
        self.fw_version_build = {'id': self.sys_id << 8 | 0x05,
                                 'type': int,
                                 'unit': None,
                                 'range': None,
                                 'default': None,
                                 'value': None}
        self.fw_version_revision = {'id': self.sys_id << 8 | 0x06,
                                    'type': int,
                                    'unit': None,
                                    'range': None,
                                    'default': None,
                                    'value': None}
        self.error_flag_register = {'id': self.sys_id << 8 | 0x07,
                                    'type': int,
                                    'unit': None,
                                    'range': None,
                                    'default': None,
                                    'value': None}
        self.proxy_fpga_version = {'id': self.sys_id << 8 | 0x08,
                                   'type': int,
                                   'unit': None,
                                   'range': None,
                                   'default': None,
                                   'value': None}
        self.cpu_fpga_version = {'id': self.sys_id << 8 | 0x09,
                                 'type': int,
                                 'unit': None,
                                 'range': None,
                                 'default': None,
                                 'value': None}
        System.__init__(self, board=board)
        self.name = self.__class__.__name__

    def GetFirmwareID(self):
        return self.get_register('firmware_id')

    def GetFirmwareBranch(self):
        return self.get_register('fw_branch')

    def GetFirmwareType(self):
        return self.get_register('fw_type')

    def GetFirmwareVersionMajor(self):
        return self.get_register('fw_version_major')

    def GetFirmwareVersionMinor(self):
        return self.get_register('fw_version_minor')

    def GetFirmwareBuild(self):
        return self.get_register('fw_version_build')

    def GetFirmwareVersionRevision(self):
        return self.get_register('fw_version_revision')

    def GetProxyFPGAVersion(self):
        return self.get_register('proxy_fpga_version')

    def GetCPUFPGAVersion(self):
        return self.get_register('cpu_fpga_version')

    def GetErrorFlagRegister(self):
        response = self.get_register('error_flag_register')
        if response[0] is not None:
            error_flag_register = parse_error_flags(response)
            if self._board.verbose:
                print(error_flag_register)
            return error_flag_register
        else:
            return response

    def ResetErrorFlagRegister(self):
        return self.set_register('error_flag_register', 0)

    def GetFirmwareSN(self, index, count):
        register = {'id': self.sys_id << 8, 'type': str}
        return self._board.get_vector(register, index, count)

    def SetFirmwareSN(self, index, vector):
        register = {'id': self.sys_id << 8, 'type': str}
        return self._board.set_vector(register, index, vector)

    def GetGitHEADSHA1Vector(self, index, count):
        register = {'id': self.sys_id << 8 | 0x01, 'type': str}
        return self._board.get_vector(register, index, count)


class BoardEEPROM(System):
    r"""
Device Functionality - BoardEEPROM Read/Write
System ID: 0x20

+--------------------+------+-----------------------------------------------------------------------------------------+
| Register Name      | Id   | Comment                                                                                 |
+====================+======+=========================================================================================+
| lock               | 0x00 | write the key to unlock, anything else to lock.Key value: 0x3f4744f6 (float 0.778396)   |
+--------------------+------+-----------------------------------------------------------------------------------------+


    """

    @staticmethod
    def help():
        print(BoardEEPROM.__doc__)

    _is_a_system = False

    def __init__(self, board=None):
        self.sys_id = 0x20
        self.lock = {'id': self.sys_id << 8 | 0x00, 'type': float, 'unit': None, 'range': None, 'default': 0.0,
                     'value': 0.778396}
        self.register = {'id': self.sys_id << 8, 'type': bytes}
        System.__init__(self, board=board)
        self.name = self.__class__.__name__

    def _SetEEPROMSegment(self, index, vector):
        resp = self._board.set_vector(self.register, index, vector)
        return resp

    def _GetEEPROMSegment(self, index, count):
        return self._board.get_vector(self.register, index, count)

    def GetEEPROM(self, index, count):
        vec = bytearray([])
        end_index = index + count
        chunk_size_in_bytes = CHUNK_SIZE * 4
        for i in range(index, index + count, chunk_size_in_bytes):
            if i + chunk_size_in_bytes > end_index:
                aux = self._GetEEPROMSegment(i, count % chunk_size_in_bytes)
            else:
                aux = self._GetEEPROMSegment(i, chunk_size_in_bytes)
            vec += aux[0]

        return vec

    def SetEEPROM(self, index, vector):
        chunk_size_in_bytes = CHUNK_SIZE * 4
        for i in range(0, len(vector), chunk_size_in_bytes):
            aux = vector[i:i + chunk_size_in_bytes]
            self._SetEEPROMSegment(i + index, aux)
        return vector

    def EEPROMLockEnable(self, enable=True):
        if enable:
            return self.set_register('lock', 0.0)
        else:
            return self.set_register('lock', self.lock['value'])

    def VerifyEEPROM(self, index, vector):
        stored_eeprom = self.GetEEPROM(index, len(vector))
        return stored_eeprom == vector, stored_eeprom


class DeviceEEPROM(System):
    r"""
Device Functionality - DeviceEEPROM Read/Write
System ID: 0x21

+--------------------+------+-----------------------------------------------------------------------------------------+
| Register Name      | Id   | Comment                                                                                 |
+====================+======+=========================================================================================+
| lock               | 0x00 | write the key to unlock, anything else to lock.Key value: 0x3edeb7aa (float 0.434995)   |
+--------------------+------+-----------------------------------------------------------------------------------------+
    """

    @staticmethod
    def help():
        print(DeviceEEPROM.__doc__)

    _is_a_system = False

    def __init__(self, board=None):
        self.sys_id = 0x21
        self.lock = {'id': self.sys_id << 8 | 0x0a, 'type': float, 'unit': None, 'range': None, 'default': 0.0,
                     'value': 0.434995}
        self.reparse = {'id': self.sys_id << 8 | 0x0a, 'type': float, 'unit': None, 'range': None, 'default': 0.0,
                        'value': 0}
        self.register = {'id': self.sys_id << 8, 'type': bytes}
        System.__init__(self, board=board, channel=None)
        self.name = self.__class__.__name__

    def _SetEEPROMSegment(self, index, vector):
        resp = self._board.set_vector(self.register, index, vector)
        return resp

    def _GetEEPROMSegment(self, index, count):
        return self._board.get_vector(self.register, index, count)

    def GetEEPROM(self, index, count):
        vec = bytearray([])
        end_index = index + count
        chunk_size_in_bytes = CHUNK_SIZE * 4
        for i in range(index, index + count, chunk_size_in_bytes):
            if i + chunk_size_in_bytes > end_index:
                aux = self._GetEEPROMSegment(i, count % chunk_size_in_bytes)
            else:
                aux = self._GetEEPROMSegment(i, chunk_size_in_bytes)
            try:
                vec += aux[0]
            except (IndexError, TypeError):
                return vec

        return vec

    def SetEEPROM(self, index, vector):
        chunk_size_in_bytes = CHUNK_SIZE * 4
        for i in range(0, len(vector), chunk_size_in_bytes):
            aux = vector[i:i + chunk_size_in_bytes]
            self._SetEEPROMSegment(i + index, aux)
        return vector

    def EEPROMLockEnable(self, enable=True):
        if enable:
            return self.set_register('lock', 0.0)
        else:
            return self.set_register('lock', self.lock['value'])

    def VerifyEEPROM(self, index, vector):
        stored_eeprom = self.GetEEPROM(index, len(vector))
        return stored_eeprom == vector, stored_eeprom

    def ReparseEEPROM(self):
        return self.get_register('reparse')


class VectorPatternMemory(System):
    """
Device Functionality - Vector Pattern Memory
System ID: 0x26

+---------------------------+------+-------------+------+-------+---------+-------------------------------------------+
| Register Name             | Id   | Type        | Unit | Range | Default | Comment                                   |
+===========================+======+=============+======+=======+=========+===========================================+
| proxy_valid               | 0x00 |float 32-bit |None  |       |         |                                           |
+---------------------------+------+-------------+------+-------+---------+-------------------------------------------+
    """

    def __init__(self, board=None):
        System.__init__(self, board=board)
        self.name = self.__class__.__name__

    sys_id = 0x26
    _readonly = False

    proxy_valid = {'id': sys_id << 8 | 0x00,
                        'type': int,
                        'unit': None,
                        'range': [0, 1],
                        'default': False,
                        'value': False}
    vector = {'id': sys_id << 8, 'type': float}


    def SetPattern(self, index, vector):
        for i in range(0, len(vector), CHUNK_SIZE):
            aux = vector[i:i + CHUNK_SIZE]
            self.SetPatternSegment(i + index, aux)

    def GetPattern(self, index, count):
        vec = []
        # end_index = index + count
        if count < CHUNK_SIZE:
            return self.GetPatternSegment(index, count)
        iters = count // CHUNK_SIZE
        remainder = count % CHUNK_SIZE
        for i in range(iters):
            aux = self.GetPatternSegment(i * CHUNK_SIZE, CHUNK_SIZE)
            vec.extend(aux)
        if remainder:
            aux = self.GetPatternSegment(index + count - remainder, remainder)
            vec.extend(aux)
        return vec

    def SetPatternSegment(self, index, vector):
        return self._board.set_vector(self.vector, index, vector)

    def GetPatternSegment(self, index: int, count: int):
        response = self._board.get_vector(self.vector, index, count)
        # return list(struct.unpack('>'+'f'*count, response))
        return response

    def GetProxyValidity(self):
        return self.get_register('proxy_valid')

    def SaveVectorMemory(self):
        return self.set_register('proxy_valid', 0)

    # def MakeChangesPersistent(self):
    #     self._board.set_value(self.pattern_segment, 0)


class SnapshotManager(System):
    """
Device Functionality - Vector Pattern Memory
System ID: 0x26

+---------------------------+------+-------------+------+-------+---------+-------------------------------------------+
| Register Name             | Id   | Type        | Unit | Range | Default | Comment                                   |
+===========================+======+=============+======+=======+=========+===========================================+
| snapshot_capacity         | 0x00 |float 32-bit |None  |       | 2       |                                           |
+---------------------------+------+-------------+------+-------+---------+-------------------------------------------+
| default_snapshot          | 0x01 |float 32-bit |None  |       | 0       |                                           |
+---------------------------+------+-------------+------+-------+---------+-------------------------------------------+
    """

    def __init__(self, board=None):
        self.sys_id = 0x27
        self._readonly = False

        self.snapshot_capacity = {'id': self.sys_id << 8 | 0x00,
                                  'type': int,
                                  'unit': None,
                                  'range': None,
                                  'default': False,
                                  'value': False}
        self.default_snapshot = {'id': self.sys_id << 8 | 0x01,
                                 'type': int,
                                 'unit': None,
                                 'range': [0, 7],
                                 'default': False,
                                 'value': False}
        System.__init__(self, board=board)
        self.name = self.__class__.__name__

    def GetSnapShotCapacity(self):
        return self.get_register('snapshot_capacity')

    def GetDefaultSnapShot(self):
        return self.get_register('default_snapshot')

    def SetDefaultSnapShot(self, value):
        return self.set_register('default_snapshot', value)


#
# class READOUT:
#     r"""
# Currently Not Implemented...
# DeviceEEPROM readout, to return the following registers records as a list:
#
# register_name               id      type
# EEPROM_VERSION                0x00    uint 32-bit
# EEPROM_SUBVERSION           0x01    uint 32-bit
# PART_CONFIGURATION          0x02    uint 32-bit
# PART_CONFIGURATION_VERSION    0x03    uint 32-bit
# MAX_POSITIVE_CURRENT        0x04    uint 32-bit
# MAX_NEGATIVE_CURRENT        0x05    uint 32-bit
# TEMPERATURE_OFFSET_A        0x06    uint 32-bit
# TEMPERATURE_GRADIENT_A        0x07    uint 32-bit
# TEMPERATURE_OFFSET_B        0x08    uint 32-bit
# TEMPERATURE_GRADIENT_B        0x09    uint 32-bit
#
#     """
#
#     @staticmethod
#     def help():
#         print(READOUT.__doc__)
#
#     sys_id = 0x21
#     _is_a_system = False
#
#     def __init__(self, channel: int = 0):
#         self._readonly = True
#
#     registers = inspect.getmembers(DeviceEEPROM, lambda a: not (inspect.isroutine(a)))
#     READOUT = [item for item in registers if
#                not (item[0].startswith('__') and item[0].endswith('__') or item[0].islower())]


# -*- coding: utf-8 -*-
from .registers import mre2_registers as Registers
from .tools.list_comports import get_mre2_port
from ._connections import Board as _Board, Channel as _Channel, ProxyBoard as _ProxyBoard
"""MR-E-2 SDK.

The MRE2 SDK is organized in the following way:

         Board.Mirror.Channel.System.RegisterCommand(parameters)

  Board:      Low/mid level object for operating the firmware (vector patterns, simple/pro commands, etc)
  Mirror:     Handles Proxy Mirror Board Operations (GetTemperature, GetConnectedStatus, etc)
  Channel:    A given channel, which can set input types, gain, etc.
  System:     Each channel has its own stages. From here can perform commands that get/set individual registers.
  registers:  Each system has several registers, which are dictionaries with elements regarding valid units/range.


To begin:
Import necessary module and initialize a board (verbose for console output)

>>>from mre2 import MRE2Board
>>>mre2 = MRE2Board(port='COM12', verbose=True)  # this will connect and synchronize all channel with the firmware

Define the system you want to interact with (board.mirror.channel.system) (see registers.system_names for a list)
>>>sig_gen = mre2.Mirror.Channel_0.SignalGenerator
Start setting individual registers
>>>sig_gen.SetAmplitude(0.1)

Additional operations are available directly through the Board object (get/set multiple, reset, etc)
For example, multiple set/get can be done. After this, the system will output a sinusoid on channel 0.
>>>mre2.set_value([sig_gen.frequency, sig_gen.amplitude, sig_gen.unit, sig_gen.run], [10.0, 0.1, Units.CURRENT, 1])
>>>mre2.get_value([sig_gen.frequency, sig_gen.amplitude, sig_gen.unit, sig_gen.run])

Example
-------
More examples to come...
"""

hwid = '0483:A31E'
"""str: default hardware id of mre2 board, for port recognition.
"""


# standard MRE2 Board
class MRE2Board(_Board):
    r"""
    Board object from which to issue MRE2 commands over a serial connection.

    Parameters
    ----------
    port, baudrate, comm_lock, timeout, verbose
        Standard serial connection parameters. Extended from Connection class.
    boot_in_simple: bool, optional = False
        Determines whether or not Board will be set in simple-mode after initialization.
    board_reset: bool, optional = False
        Determines whether or not to reset the Board during initialization.

    Raises
    ------
    none

    Notes
    -----
    Board extends Connection class. Refer to Connection for methods, parameters, etc.
    Board inherits Commands. Refer to Commands for methods, parameters, etc.

    Examples
    --------
    Instantiate a simple-mode Board with connection over COM12 port, with baud rate 9600. Then switch into pro-mode.
    This will not flush the firmware registers, so it is assumed to already be reset.

    >>> board = MRE2Board(port='COM12', baudrate=9600, verbose=True, simple=True)
    Establishing Connection via port [COM12]...
    Connected.
    Initializing Board in Simple Mode...
    Sending: b'~\x00\x06\x02\x00\x00\x00\x00~'
    Received: b''
    Verifying Simple via Handshake...
    Sending: b'Start\r\n'
    Received: b'NO\r\n'
    Handshake NOT Successful: UNKNOWN_ERROR_CODE.
    Sending: b'Start\r\n'
    Received: b'OK\r\n'
    Handshake Successful.
    Board Initialized in Simple Mode.
    Initialization Finalized. Board is Ready.
    >>> board.go_pro()
    Sending: b'GoPro\r\n'
    Received: b'OK\r\n'
    Go Pro Successful.

    """
    def __init__(self, port: str = None, baudrate: int = 115200, comm_lock: bool = True,
                 timeout: float = 2, verbose: bool = False, boot_in_simple: bool = False,
                 board_reset: bool = False, inter_byte_timeout: float = 0.2):
        if port is None:
            try:
                port = get_mre2_port()
            except IndexError:
                print("MRE2 Device Not Found. Check Connection.")

        _Board.__init__(self, port=port, baudrate=baudrate, comm_lock=comm_lock,
                        timeout=timeout, verbose=verbose, simple=boot_in_simple,
                        board_reset=board_reset, inter_byte_timeout=inter_byte_timeout)

        # board-level systems
        self.VectorPatternMemory = Registers.VectorPatternMemory(board=self)
        self.TemperatureManager = Registers.TemperatureManager(board=self)
        self.MISC = Registers.Misc(board=self)
        self.Logger = Registers.Logger(board=self)
        # mirror proxy board
        del self.ProxyBoard
        self.Mirror = MRE2Mirror(board=self)
        self.channel = [self.Mirror.Channel_0, self.Mirror.Channel_1]
        self.systems = Registers.systems()
        if self.verbose:
            print('Initialization Finalized, Board is Ready.')


class MRE2Mirror(_ProxyBoard):
    r"""
    Mirror object which contains mirror-level systems and commands. Such as GetProxyTemperature, GetConnectedStatus, etc

    Parameters
    ----------
    board
        Handle for the Board object

    Raises
    ------
    none

    Notes
    -----
    MRE2Mirror is a subclass of ProxyBoard. Refer to optoKummenberg.connections for methods, parameters, etc.

    Examples
    --------
    Instantiate a simple-mode Board with connection over COM12 port, with baud rate 9600. Then switch into pro-mode.
    This will not flush the firmware registers, so it is assumed to already be reset.

    >>> import optoMDC
    >>> mre2 = optoMDC.connect()
    >>> mre2.Mirror.GetConnectedStatus()

    """
    def __init__(self, board):
        _ProxyBoard.__init__(self, board)
        self.EEPROM = Registers.DeviceEEPROM(board)

        # channel-level systems
        self.Channel_0 = MRE2Channel(board, 0)
        self.Channel_1 = MRE2Channel(board, 1)

        # proxy board systems
        self.OpticalFeedback = Registers.OpticalFeedback(board)


class MRE2Channel(_Channel):
    r"""
    Channel object which contains channel-level systems and commands. Such as StaticInput, SetControlMode, etc

    Parameters
    ----------
    board
        Handle for the Board object
    channel_number
        integer of the channel (0 or 1 by default)

    Raises
    ------
    none

    Notes
    -----
    MRE2Channel is an subclass of Channel. Refer to optoKummenberg.connections for methods, parameters, etc.

    Examples
    --------
    Instantiate a simple-mode Board with connection over COM12 port, with baud rate 9600. Then switch into pro-mode.
    This will not flush the firmware registers, so it is assumed to already be reset.

    >>> import optoMDC
    >>> mre2 = optoMDC.connect()
    >>> mre2.Mirror.Channel_0.GetControlMode()
    >>> mre2.Mirror.Channel_1.SignalGenerator.GetAmplitude()

    """
    def __init__(self, board, channel_number: int = 0):
        _Channel.__init__(self, board, channel_number)

        # meta systems
        self.VectorPatternUnit = Registers.VectorPatternUnit(self._channel, self._board)
        self.VectorPatternUnit.Memory = self._board.VectorPatternMemory
        # inputs
        self.Analog            = Registers.Analog(self._channel, self._board)
        self.SignalGenerator   = Registers.SignalGenerator(self._channel, self._board)

        # control modes
        self.OFPID             = Registers.OFPID(self._channel, self._board)
        self.XYPID             = Registers.XYPID(self._channel, self._board)



from phootonics_controller.utils.MR_E_2_Python_SDK.optoKummenberg.tools.list_comports import *


def get_mre2_port(hwid='0483:A31E'):
    """str: default hardware id of mre2 board, for port recognition.
    """
    devices = list_ports.comports()
    device = list([device.device for device in devices if hwid in device.hwid])
    return device.pop() if device else None



# /*****************************************************************************
# * | File        :	  config.py
# * | Author      :   Waveshare team
# * | Function    :   Hardware underlying interface
# * | Info        :
# *----------------
# * | This version:   V1.0
# * | Date        :   2020-12-12
# * | Info        :
# ******************************************************************************/
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documnetation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to  whom the Software is
# furished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS OR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import spidev
import RPi.GPIO as GPIO
import time

# Pin definition
RST_PIN = 18
CS_PIN = 22
DRDY_PIN = 17

# SPI device, bus = 0, device = 0
SPI = spidev.SpiDev(0, 0)


def digital_write(pin, value):
    GPIO.output(pin, value)


def digital_read(pin):
    return GPIO.input(DRDY_PIN)


def delay_ms(delaytime):
    time.sleep(delaytime / 1000.0)


def spi_writebyte(data):
    SPI.writebytes(data)


def spi_readbytes(reg):
    return SPI.readbytes(reg)


def module_init():
    GPIO.setmode(GPIO.BCM)
    GPIO.setwarnings(False)
    GPIO.setup(RST_PIN, GPIO.OUT)
    GPIO.setup(CS_PIN, GPIO.OUT)
    # GPIO.setup(DRDY_PIN, GPIO.IN)
    GPIO.setup(DRDY_PIN, GPIO.IN, pull_up_down=GPIO.PUD_UP)
    SPI.max_speed_hz = 200000
    SPI.mode = 0b01
    return 0


def module_exit():
    SPI.close()
    GPIO.output(RST_PIN, 0)
    GPIO.output(CS_PIN, 0)
### END OF FILE ###